# -*- coding: utf-8 -*-
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Additional help about CRC32C and installing crcmod."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from gslib.help_provider import HelpProvider

_DETAILED_HELP_TEXT = ("""
<B>OVERVIEW</B>
  Google Cloud Storage provides a cyclic redundancy check (CRC) header that
  allows clients to verify the integrity of object contents. For non-composite
  objects Google Cloud Storage also provides an MD5 header to allow clients to
  verify object integrity, but for composite objects only the CRC is available.
  gsutil automatically performs integrity checks on all uploads and downloads.
  Additionally, you can use the ``gsutil hash`` command to calculate a CRC for
  any local file.

  The CRC variant used by Google Cloud Storage is called CRC32C (Castagnoli),
  which is not available in the standard Python distribution. The implementation
  of CRC32C used by gsutil is provided by a third-party Python module called
  `crcmod <https://pypi.python.org/pypi/crcmod>`_.

  The crcmod module contains a pure-Python implementation of CRC32C, but using
  it results in very poor performance. A Python C extension is also provided by
  crcmod, which requires compiling into a binary module for use. gsutil ships
  with a precompiled crcmod C extension for macOS; for other platforms, see
  the installation instructions below.

  At the end of each copy operation, the ``gsutil cp`` and ``gsutil rsync``
  commands validate that the checksum of the source file/object matches the
  checksum of the destination file/object. If the checksums do not match,
  gsutil will delete the invalid copy and print a warning message. This very
  rarely happens, but if it does, please contact gs-team@google.com.


<B>CONFIGURATION</B>
  To determine if the compiled version of crcmod is available in your Python
  environment, you can inspect the output of the ``gsutil version`` command for
  the "compiled crcmod" entry:

    $ gsutil version -l
    ...
    compiled crcmod: True
    ...

  If your crcmod library is compiled to a native binary, this value will be
  True. If using the pure-Python version, the value will be False.

  To control gsutil's behavior in response to crcmod's status, you can set the
  "check_hashes" configuration variable. For details on this variable, see the
  surrounding comments in your boto configuration file. If "check_hashes"
  is not present in your configuration file, rerun ``gsutil config`` to
  regenerate the file.


<B>INSTALLATION</B>
  These installation instructions assume that:

  -  You have ``pip`` installed. Consult the `pip installation instructions
     <https://pip.pypa.io/en/stable/installing/>`_ for details on how
     to install ``pip``.
  -  Your installation of ``pip`` can be found in your ``PATH`` environment
     variable. If it cannot, you may need to replace ``pip`` in the commands
     below with the full path to the executable.
  -  You are installing the crcmod package for use with your system installation
     of Python, and thus use the ``sudo`` command. If installing crcmod for a
     different Python environment (e.g. in a virtualenv), you should omit
     ``sudo`` from the commands below.

  CentOS, RHEL, and Fedora
  ------------------------

  Note that CentOS 6 and similar variants use Python 2.6 by default, which will
  not run gsutil. To enable Python 2.7 and compile/install crcmod on CentOS 6:

    sudo su  # Run as root; need shell session with Python 2.7 enabled
    yum install gcc python-devel python-setuptools redhat-rpm-config
    source /opt/rh/python27/enable  # Make default `python` executable use 2.7.X
    python -m pip install -U pip  # Upgrade old default version of pip
    python -m pip uninstall crcmod
    python -m pip install --no-cache-dir -U crcmod
    exit  # Exit su session

  To compile and install crcmod on OS versions that use Python 2.7 by default:

    sudo yum install gcc python-devel python-setuptools redhat-rpm-config
    sudo pip uninstall crcmod
    sudo pip install --no-cache-dir -U crcmod

  Debian and Ubuntu
  -----------------

  To compile and install crcmod:

    sudo apt-get install gcc python-dev python-setuptools
    sudo pip uninstall crcmod
    sudo pip install --no-cache-dir -U crcmod

  Enterprise SUSE
  -----------------

  To compile and install crcmod:

    sudo zypper install gcc python-devel
    sudo pip uninstall crcmod
    sudo pip install --no-cache-dir -U crcmod

  macOS
  -----

  gsutil distributes a pre-compiled version of crcmod for macOS, so you shouldn't
  need to compile and install it yourself. If for some reason the pre-compiled
  version is not being detected, please let the Google Cloud Storage team know
  (see ``gsutil help support``).

  To compile manually on macOS, you will first need to install
  `XCode <https://developer.apple.com/xcode/>`_ and then run:

    sudo pip install -U crcmod

  Windows
  -------

  An installer is available for the compiled version of crcmod from the Python
  Package Index (PyPi) at the following URL:

  https://pypi.python.org/pypi/crcmod/1.7

  MSI installers are available for the 32-bit versions of Python 2.7.
  Make sure to install to a 32-bit Python directory. If you're using 64-bit
  Python it won't work with 32-bit crcmod, and instead you'll need to install
  32-bit Python in order to use crcmod.

  NOTE: If you have installed crcmod and gsutil hasn't detected it, it may have
  been installed to the wrong directory. It should be located at
  <python_dir>\\files\\Lib\\site-packages\\crcmod\\

  In some cases the installer will incorrectly install to
  <python_dir>\\Lib\\site-packages\\crcmod\\

  Manually copying the crcmod directory to the correct location should resolve
  the issue.
""")


class CommandOptions(HelpProvider):
  """Additional help about CRC32C and installing crcmod."""

  # Help specification. See help_provider.py for documentation.
  help_spec = HelpProvider.HelpSpec(
      help_name='crc32c',
      help_name_aliases=['crc32', 'crc', 'crcmod'],
      help_type='additional_help',
      help_one_line_summary='CRC32C and Installing crcmod',
      help_text=_DETAILED_HELP_TEXT,
      subcommand_help_text={},
  )



# -*- coding: utf-8 -*-
# Copyright 2019 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Integration tests for the hmac command."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re
import boto

from gslib.project_id import PopulateProjectId
import gslib.tests.testcase as testcase
from gslib.tests.testcase.integration_testcase import SkipForS3
from gslib.tests.testcase.integration_testcase import SkipForXML
from gslib.tests.util import SetBotoConfigForTest
from gslib.tests.util import unittest
from gslib.utils.retry_util import Retry


def _LoadServiceAccount(account_field):
  return boto.config.get_value('GSUtil', account_field)


SERVICE_ACCOUNT = _LoadServiceAccount('test_hmac_service_account')
ALT_SERVICE_ACCOUNT = _LoadServiceAccount('test_hmac_alt_service_account')
LIST_SERVICE_ACCOUNT = _LoadServiceAccount('test_hmac_list_service_account')

MAX_SA_HMAC_KEYS = 5


class KeyLimitError(Exception):
  pass


@SkipForS3('S3 does not have an equivalent API')
@SkipForXML('XML HMAC control is not supported.')
class TestHmacIntegration(testcase.GsUtilIntegrationTestCase):
  """Hmac integration test cases.

  These tests rely on the presence of 3 service accounts specified in the BOTO
  config. test_hmac_service_account and test_hmac_alt_service_account should not
  have any undeleted keys and test_hmac_list_service_account should have only
  deleted and active keys.
  """

  def ExtractAccessId(self, output_string):
    id_match = re.search(r'(GOOG[\S]*)', output_string)
    if not id_match:
      self.fail('Couldn\'t find Access Id in output string:\n"%s"' %
                output_string)
    return id_match.group(0)

  def ExtractEtag(self, output_string):
    etag_match = re.search(r'\sEtag:\s+([\S]+)$', output_string)
    if not etag_match:
      self.fail('Couldn\'t find Etag in output string:\n"%s"' % output_string)
    return etag_match.group(1)

  def AssertKeyMetadataMatches(self,
                               output_string,
                               access_id='GOOG.*',
                               state='ACTIVE',
                               service_account='.*',
                               project='.*'):
    self.assertRegexpMatches(output_string, r'Access ID %s:' % access_id)
    self.assertRegexpMatches(output_string, r'\sState:\s+%s' % state)
    self.assertRegexpMatches(output_string,
                             r'\s+Service Account:\s+%s\n' % service_account)
    self.assertRegexpMatches(output_string, r'\s+Project:\s+%s' % project)
    self.assertRegexpMatches(output_string, r'\s+Time Created:\s+.*')
    self.assertRegexpMatches(output_string, r'\s+Time Last Updated:\s+.*')

  def CleanupHelper(self, access_id):
    # Set the key to inactive if it isn't already.
    try:
      self.RunGsUtil(['hmac', 'update', '-s', 'INACTIVE', access_id])
    except AssertionError as e:
      if 'Update must modify the credential' not in str(e):
        raise

    self.RunGsUtil(['hmac', 'delete', access_id])

  @Retry(KeyLimitError, tries=5, timeout_secs=3)
  def _CreateWithRetry(self, service_account):
    """Retry creation on key limit failures."""

    try:
      return self.RunGsUtil(['hmac', 'create', service_account],
                            return_stdout=True)
    except AssertionError as e:
      if 'HMAC key limit reached' in str(e):
        raise KeyLimitError(str(e))
      else:
        raise

  def CreateHelper(self, service_account):
    stdout = self._CreateWithRetry(service_account)
    return self.ExtractAccessId(stdout)

  def test_malformed_commands(self):
    params = [
        ('hmac create', 'requires a service account'),
        ('hmac create -u email', 'requires a service account'),
        ('hmac create -p proj', 'requires a service account'),
        ('hmac delete', 'requires an Access ID'),
        ('hmac delete -p proj', 'requires an Access ID'),
        ('hmac get', 'requires an Access ID'),
        ('hmac get -p proj', 'requires an Access ID'),
        ('hmac list account1', 'unexpected arguments'),
        ('hmac update keyname', 'state flag must be supplied'),
        ('hmac update -s KENTUCKY', 'state flag value must be one of'),
        ('hmac update -s INACTIVE', 'requires an Access ID'),
        ('hmac update -s INACTIVE -p proj', 'requires an Access ID'),
    ]
    for command, error_substr in params:
      stderr = self.RunGsUtil(command.split(),
                              return_stderr=True,
                              expected_status=1)
      self.assertIn(error_substr, stderr)

  @unittest.skipUnless(SERVICE_ACCOUNT,
                       'Test requires service account configuration.')
  def test_create(self):
    stdout = self.RunGsUtil(['hmac', 'create', SERVICE_ACCOUNT],
                            return_stdout=True)
    try:
      self.assertRegexpMatches(stdout, r'Access ID:\s+\S+')
      self.assertRegexpMatches(stdout, r'Secret:\s+\S+')
    finally:
      access_id = self.ExtractAccessId(stdout)
      self.CleanupHelper(access_id)

  def test_create_sa_not_found(self):
    stderr = self.RunGsUtil(['hmac', 'create', 'DNE@mail.com'],
                            return_stderr=True,
                            expected_status=1)
    self.assertIn('Service Account \'DNE@mail.com\' not found.', stderr)

  @unittest.skipUnless(ALT_SERVICE_ACCOUNT,
                       'Test requires service account configuration.')
  def test_delete(self):
    access_id = self.CreateHelper(ALT_SERVICE_ACCOUNT)
    self.RunGsUtil(['hmac', 'update', '-s', 'INACTIVE', access_id])

    self.RunGsUtil(['hmac', 'delete', access_id])

    stdout = self.RunGsUtil(['hmac', 'list', '-u', ALT_SERVICE_ACCOUNT],
                            return_stdout=True)
    self.assertNotIn(access_id, stdout)

  @unittest.skipUnless(SERVICE_ACCOUNT,
                       'Test requires service account configuration.')
  def test_delete_active_key(self):
    access_id = self.CreateHelper(SERVICE_ACCOUNT)

    stderr = self.RunGsUtil(['hmac', 'delete', access_id],
                            return_stderr=True,
                            expected_status=1)

    try:
      self.assertIn('400 Cannot delete keys in ACTIVE state.', stderr)
    finally:
      self.CleanupHelper(access_id)

  def test_delete_key_not_found(self):
    stderr = self.RunGsUtil(['hmac', 'delete', 'GOOG1234DNE'],
                            return_stderr=True,
                            expected_status=1)

    self.assertIn('404 Access ID not found', stderr)

  @unittest.skipUnless(ALT_SERVICE_ACCOUNT,
                       'Test requires service account configuration.')
  def test_get(self):
    access_id = self.CreateHelper(ALT_SERVICE_ACCOUNT)
    stdout = self.RunGsUtil(['hmac', 'get', access_id], return_stdout=True)

    try:
      self.AssertKeyMetadataMatches(
          stdout,
          access_id=access_id,
          service_account=ALT_SERVICE_ACCOUNT,
          project=PopulateProjectId(None),
      )
    finally:
      self.CleanupHelper(access_id)

  def test_get_not_found(self):
    stderr = self.RunGsUtil(['hmac', 'get', 'GOOG1234DNE'],
                            return_stderr=True,
                            expected_status=1)

    self.assertIn('404 Access ID not found', stderr)

  def setUpListTest(self):
    for _ in range(MAX_SA_HMAC_KEYS):
      self.RunGsUtil(['hmac', 'create', LIST_SERVICE_ACCOUNT],
                     expected_status=None)

  @unittest.skipUnless(LIST_SERVICE_ACCOUNT and SERVICE_ACCOUNT,
                       'Test requires service account configuration.')
  def test_list(self):
    # Ensure LIST_SERVICE_ACCOUNT has MAX_SA_HMAC_KEYS. These keys should all be
    # active otherwise the tests will fail.
    self.setUpListTest()
    alt_access_id = self.CreateHelper(SERVICE_ACCOUNT)
    self.RunGsUtil(['hmac', 'update', '-s', 'INACTIVE', alt_access_id])

    # Test listing single service account only - non deleted keys
    stdout = self.RunGsUtil(['hmac', 'list', '-u', LIST_SERVICE_ACCOUNT],
                            return_stdout=True)
    stdout = stdout.strip().split('\n')

    list_account_key_count = 0
    for line in stdout:
      access_id, state, account = line.split()
      self.assertIn('GOOG', access_id)
      self.assertEqual(account, LIST_SERVICE_ACCOUNT)
      self.assertEqual(state, 'ACTIVE')
      list_account_key_count += 1
    self.assertEqual(list_account_key_count, MAX_SA_HMAC_KEYS)

    # Test listing all project service account HMAC keys.
    stdout = self.RunGsUtil(['hmac', 'list'], return_stdout=True)
    stdout = stdout.strip().split('\n')

    project_key_count = 0
    inactive_key_listed = False
    for line in stdout:
      _, state, account = line.split()
      project_key_count += 1
      if account == SERVICE_ACCOUNT and state == 'INACTIVE':
        inactive_key_listed = True
    # Other tests generate keys for this project. Can't assert exact quantity.
    self.assertGreater(project_key_count, list_account_key_count)
    self.assertTrue(inactive_key_listed)

    # Test listing deleted keys.
    self.RunGsUtil(['hmac', 'delete', alt_access_id])

    stdout = self.RunGsUtil(['hmac', 'list', '-a'], return_stdout=True)
    stdout = stdout.strip().split('\n')

    project_key_count = 0
    deleted_key_listed = False
    for line in stdout:
      _, state, account = line.split()
      project_key_count += 1
      if account == SERVICE_ACCOUNT and state == 'DELETED':
        deleted_key_listed = True
    self.assertTrue(deleted_key_listed)
    self.assertGreater(project_key_count, list_account_key_count)

  def ParseListOutput(self, stdout):
    current_key = ''
    for l in stdout.split('\n'):
      if current_key and l.startswith('Access ID'):
        yield current_key
        current_key = l
      else:
        current_key += l
      current_key += '\n'

  @unittest.skipUnless(LIST_SERVICE_ACCOUNT and ALT_SERVICE_ACCOUNT,
                       'Test requires service account configuration.')
  def test_list_long_format(self):
    # Ensure LIST_SERVICE_ACCOUNT has MAX_SA_HMAC_KEYS. These keys should all be
    # active otherwise the tests will fail.
    self.setUpListTest()
    alt_access_id = self.CreateHelper(ALT_SERVICE_ACCOUNT)
    self.RunGsUtil(['hmac', 'update', '-s', 'INACTIVE', alt_access_id])

    stdout = self.RunGsUtil(['hmac', 'list', '-l'], return_stdout=True)

    try:
      self.assertIn(' ACTIVE', stdout)
      self.assertIn('INACTIVE', stdout)
      self.assertIn(ALT_SERVICE_ACCOUNT, stdout)
      self.assertIn(LIST_SERVICE_ACCOUNT, stdout)
      for key_metadata in self.ParseListOutput(stdout):
        self.AssertKeyMetadataMatches(key_metadata, state='.*')
    finally:
      self.CleanupHelper(alt_access_id)

  def test_list_service_account_not_found(self):
    service_account = 'service-account-DNE@gmail.com'
    stderr = self.RunGsUtil(['hmac', 'list', '-u', service_account],
                            return_stderr=True,
                            expected_status=1)

    self.assertIn('Service Account \'%s\' not found.' % service_account, stderr)

  @unittest.skipUnless(ALT_SERVICE_ACCOUNT,
                       'Test requires service account configuration.')
  def test_update(self):
    access_id = self.CreateHelper(ALT_SERVICE_ACCOUNT)

    stdout = self.RunGsUtil(['hmac', 'get', access_id], return_stdout=True)
    etag = self.ExtractEtag(stdout)

    try:
      self.AssertKeyMetadataMatches(stdout, state='ACTIVE')

      stdout = self.RunGsUtil(
          ['hmac', 'update', '-s', 'INACTIVE', '-e', etag, access_id],
          return_stdout=True)
      self.AssertKeyMetadataMatches(stdout, state='INACTIVE')
      stdout = self.RunGsUtil(['hmac', 'get', access_id], return_stdout=True)
      self.AssertKeyMetadataMatches(stdout, state='INACTIVE')

      stdout = self.RunGsUtil(['hmac', 'update', '-s', 'ACTIVE', access_id],
                              return_stdout=True)
      self.AssertKeyMetadataMatches(stdout, state='ACTIVE')
      stdout = self.RunGsUtil(['hmac', 'get', access_id], return_stdout=True)
      self.AssertKeyMetadataMatches(stdout, state='ACTIVE')

      stderr = self.RunGsUtil(
          ['hmac', 'update', '-s', 'INACTIVE', '-e', 'badEtag', access_id],
          return_stderr=True,
          expected_status=1)
      self.assertIn('Etag does not match expected value.', stderr)
    finally:
      self.CleanupHelper(access_id)


@SkipForS3('S3 does not have an equivalent API')
class TestHmacXmlIntegration(testcase.GsUtilIntegrationTestCase):
  """XML integration tests for the "hmac" command."""
  boto_config_hmac_auth_only = [
      # Overwrite other credential types.
      ('Credentials', 'gs_oauth2_refresh_token', None),
      ('Credentials', 'gs_service_client_id', None),
      ('Credentials', 'gs_service_key_file', None),
      ('Credentials', 'gs_service_key_file_password', None),
      # Add hmac credentials.
      ('Credentials', 'gs_access_key_id', 'dummykey'),
      ('Credentials', 'gs_secret_access_key', 'dummysecret'),
  ]

  def test_hmac_fails_for_xml(self):
    with SetBotoConfigForTest(self.boto_config_hmac_auth_only):
      for subcommand in ['create', 'delete', 'get', 'list', 'update']:
        command = ['hmac', subcommand]
        stderr = self.RunGsUtil(command, expected_status=1, return_stderr=True)
        self.assertIn(
            'The "hmac" command can only be used with the GCS JSON API', stderr)



# -*- coding: utf-8 -*-
# Copyright 2012 Google Inc. All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish, dis-
# tribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the fol-
# lowing conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABIL-
# ITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT
# SHALL THE AUTHOR BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
# WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
"""Unit tests for PluralityCheckableIterator."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import sys

import six
from six.moves import range

from gslib.plurality_checkable_iterator import PluralityCheckableIterator
import gslib.tests.testcase as testcase


class CustomTestException(Exception):
  pass


class PluralityCheckableIteratorTests(testcase.GsUtilUnitTestCase):
  """Unit tests for PluralityCheckableIterator."""

  def testPluralityCheckableIteratorWith0Elems(self):
    """Tests empty PluralityCheckableIterator."""
    input_list = list(range(0))
    it = iter(input_list)
    pcit = PluralityCheckableIterator(it)
    self.assertTrue(pcit.IsEmpty())
    self.assertFalse(pcit.HasPlurality())
    output_list = list(pcit)
    self.assertEqual(input_list, output_list)

  def testPluralityCheckableIteratorWith1Elem(self):
    """Tests PluralityCheckableIterator with 1 element."""
    input_list = list(range(1))
    it = iter(input_list)
    pcit = PluralityCheckableIterator(it)
    self.assertFalse(pcit.IsEmpty())
    self.assertFalse(pcit.HasPlurality())
    output_list = list(pcit)
    self.assertEqual(input_list, output_list)

  def testPluralityCheckableIteratorWith2Elems(self):
    """Tests PluralityCheckableIterator with 2 elements."""
    input_list = list(range(2))
    it = iter(input_list)
    pcit = PluralityCheckableIterator(it)
    self.assertFalse(pcit.IsEmpty())
    self.assertTrue(pcit.HasPlurality())
    output_list = list(pcit)
    self.assertEqual(input_list, output_list)

  def testPluralityCheckableIteratorWith3Elems(self):
    """Tests PluralityCheckableIterator with 3 elements."""
    input_list = list(range(3))
    it = iter(input_list)
    pcit = PluralityCheckableIterator(it)
    self.assertFalse(pcit.IsEmpty())
    self.assertTrue(pcit.HasPlurality())
    output_list = list(pcit)
    self.assertEqual(input_list, output_list)

  def testPluralityCheckableIteratorWith1Elem1Exception(self):
    """Tests PluralityCheckableIterator with 2 elements.

    The second element raises an exception.
    """

    class IterTest(six.Iterator):

      def __init__(self):
        self.position = 0

      def __iter__(self):
        return self

      def __next__(self):
        if self.position == 0:
          self.position += 1
          return 1
        elif self.position == 1:
          self.position += 1
          raise CustomTestException('Test exception')
        else:
          raise StopIteration()

    pcit = PluralityCheckableIterator(IterTest())
    self.assertFalse(pcit.IsEmpty())
    self.assertTrue(pcit.HasPlurality())
    iterated_value = None
    try:
      for value in pcit:
        iterated_value = value
      self.fail('Expected exception from iterator')
    except CustomTestException:
      pass
    self.assertEqual(iterated_value, 1)

  def testPluralityCheckableIteratorWith2Exceptions(self):
    """Tests PluralityCheckableIterator with 2 elements that both raise."""

    class IterTest(six.Iterator):

      def __init__(self):
        self.position = 0

      def __iter__(self):
        return self

      def __next__(self):
        if self.position < 2:
          self.position += 1
          raise CustomTestException('Test exception %s' % self.position)
        else:
          raise StopIteration()

    pcit = PluralityCheckableIterator(IterTest())
    try:
      pcit.PeekException()
      self.fail('Expected exception 1 from PeekException')
    except CustomTestException as e:
      self.assertIn(str(e), 'Test exception 1')
    try:
      for _ in pcit:
        pass
      self.fail('Expected exception 1 from iterator')
    except CustomTestException as e:
      self.assertIn(str(e), 'Test exception 1')
    try:
      pcit.PeekException()
      self.fail('Expected exception 2 from PeekException')
    except CustomTestException as e:
      self.assertIn(str(e), 'Test exception 2')
    try:
      for _ in pcit:
        pass
      self.fail('Expected exception 2 from iterator')
    except CustomTestException as e:
      self.assertIn(str(e), 'Test exception 2')
    for _ in pcit:
      self.fail('Expected StopIteration')

  def testPluralityCheckableIteratorWithYieldedException(self):
    """Tests PCI with an iterator that yields an exception.

    The yielded exception is in the form of a tuple and must also contain a
    stack trace.
    """

    class IterTest(six.Iterator):

      def __init__(self):
        self.position = 0

      def __iter__(self):
        return self

      def __next__(self):
        if self.position == 0:
          try:
            self.position += 1
            raise CustomTestException('Test exception 0')
          except CustomTestException as e:
            return (e, sys.exc_info()[2])
        elif self.position == 1:
          self.position += 1
          return 1
        else:
          raise StopIteration()

    pcit = PluralityCheckableIterator(IterTest())
    iterated_value = None
    try:
      for _ in pcit:
        pass
      self.fail('Expected exception 0 from iterator')
    except CustomTestException as e:
      self.assertIn(str(e), 'Test exception 0')
    for value in pcit:
      iterated_value = value
    self.assertEqual(iterated_value, 1)

  def testPluralityCheckableIteratorReadsAheadAsNeeded(self):
    """Tests that the PCI does not unnecessarily read new elements."""

    class IterTest(six.Iterator):

      def __init__(self):
        self.position = 0

      def __iter__(self):
        return self

      def __next__(self):
        if self.position == 3:
          raise StopIteration()
        self.position += 1

    # IsEmpty and PeekException should retrieve only 1 element from the
    # underlying iterator.
    pcit = PluralityCheckableIterator(IterTest())
    pcit.IsEmpty()
    pcit.PeekException()
    self.assertEquals(pcit.orig_iterator.position, 1)
    # HasPlurality requires populating 2 elements into the iterator.
    pcit.HasPlurality()
    self.assertEquals(pcit.orig_iterator.position, 2)
    # next should yield already-populated elements without advancing the
    # iterator.
    next(pcit)  # Yields element 1
    self.assertEquals(pcit.orig_iterator.position, 2)
    next(pcit)  # Yields element 2
    self.assertEquals(pcit.orig_iterator.position, 2)
    next(pcit)  # Yields element 3
    self.assertEquals(pcit.orig_iterator.position, 3)
    try:
      next(pcit)  # Underlying iterator is empty
      self.fail('Expected StopIteration')
    except StopIteration:
      pass



# -*- coding: utf-8 -*-
# Copyright 2018 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides a tzinfo subclass to represent the UTC timezone."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from datetime import timedelta
from datetime import tzinfo


class UTC(tzinfo):
  """Timezone information class used to convert datetime timestamps to UTC.

  This class is necessary to convert timestamps to UTC. By default Python
  datetime objects are timezone unaware. This created problems when interacting
  with cloud object timestamps which are timezone-aware. This issue appeared
  when handling the timeCreated metadata attribute; the values returned by the
  service were placed in RFC 3339 format in the storage_v1_messages module. RFC
  3339 requires a timezone in any timestamp. This caused problems as the
  datetime object elsewhere in the code was timezone unaware and was different
  by exactly one hour. The main problem was that the local system used daylight
  savings time which adjusted the timestamp ahead by one hour.
  """

  def utcoffset(self, _):
    """An offset of the number of minutes away from UTC this tzinfo object is.

    Returns:
      A time duration of zero. UTC is zero minutes away from UTC.
    """
    return timedelta(0)

  def tzname(self, _):
    """A method to retrieve the name of this timezone object.

    Returns:
      The name of the timezone (i.e. 'UTC').
    """
    return 'UTC'

  def dst(self, _):
    """A fixed offset to handle daylight savings time (DST).

    Returns:
      A time duration of zero as UTC does not use DST.
    """
    return timedelta(0)



# -*- coding: utf-8 -*-
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Integration tests for retention command."""

from __future__ import absolute_import

import datetime
import re

import gslib.tests.testcase as testcase
from gslib.tests.testcase.integration_testcase import SkipForS3
from gslib.tests.testcase.integration_testcase import SkipForXML
from gslib.tests.util import ObjectToURI as suri

_SECONDS_IN_DAY = 24 * 60 * 60
_DAYS_IN_MONTH = 31
_SECONDS_IN_MONTH = _DAYS_IN_MONTH * _SECONDS_IN_DAY
_DAYS_IN_YEAR = 365.25
_SECONDS_IN_YEAR = int(_DAYS_IN_YEAR * _SECONDS_IN_DAY)


class TestRetention(testcase.GsUtilIntegrationTestCase):
  """Integration tests for retention command."""

  @SkipForS3('Retention is not supported for s3 objects.')
  @SkipForXML('Retention is not supported for XML API.')
  def test_set_retention_seconds(self):
    bucket_uri = self.CreateBucket()
    self.RunGsUtil(['retention', 'set', '60s', suri(bucket_uri)])
    self.VerifyRetentionPolicy(bucket_uri,
                               expected_retention_period_in_seconds=60)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_set_retention_days(self):
    bucket_uri = self.CreateBucket()
    self.RunGsUtil(['retention', 'set', '1d', suri(bucket_uri)])
    self.VerifyRetentionPolicy(
        bucket_uri, expected_retention_period_in_seconds=_SECONDS_IN_DAY)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_set_retention_months(self):
    bucket_uri = self.CreateBucket()
    self.RunGsUtil(['retention', 'set', '1m', suri(bucket_uri)])
    self.VerifyRetentionPolicy(
        bucket_uri, expected_retention_period_in_seconds=_SECONDS_IN_MONTH)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_set_retention_years(self):
    bucket_uri = self.CreateBucket()
    self.RunGsUtil(['retention', 'set', '1y', suri(bucket_uri)])
    self.VerifyRetentionPolicy(
        bucket_uri, expected_retention_period_in_seconds=_SECONDS_IN_YEAR)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_set_retention_multiple_sequential(self):
    bucket1_uri = self.CreateBucket()
    bucket2_uri = self.CreateBucket()

    self.RunGsUtil(
        ['retention', 'set', '1y',
         suri(bucket1_uri),
         suri(bucket2_uri)])

    self.VerifyRetentionPolicy(
        bucket1_uri, expected_retention_period_in_seconds=_SECONDS_IN_YEAR)
    self.VerifyRetentionPolicy(
        bucket2_uri, expected_retention_period_in_seconds=_SECONDS_IN_YEAR)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_set_retention_multiple_parallel(self):
    bucket1_uri = self.CreateBucket()
    bucket2_uri = self.CreateBucket()

    self.RunGsUtil(
        ['-m', 'retention', 'set', '1y',
         suri(bucket1_uri),
         suri(bucket2_uri)])

    self.VerifyRetentionPolicy(
        bucket1_uri, expected_retention_period_in_seconds=_SECONDS_IN_YEAR)
    self.VerifyRetentionPolicy(
        bucket2_uri, expected_retention_period_in_seconds=_SECONDS_IN_YEAR)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_increase_retention_unlocked(self):
    bucket_uri = self.CreateBucketWithRetentionPolicy(
        retention_period_in_seconds=_SECONDS_IN_DAY)
    self.RunGsUtil(['retention', 'set', '1m', suri(bucket_uri)])
    self.VerifyRetentionPolicy(
        bucket_uri, expected_retention_period_in_seconds=_SECONDS_IN_MONTH)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_decrease_retention_unlocked(self):
    bucket_uri = self.CreateBucketWithRetentionPolicy(
        retention_period_in_seconds=_SECONDS_IN_MONTH)
    self.RunGsUtil(
        ['retention', 'set', '{}s'.format(_SECONDS_IN_DAY),
         suri(bucket_uri)])
    self.VerifyRetentionPolicy(
        bucket_uri, expected_retention_period_in_seconds=_SECONDS_IN_DAY)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_clear_unlocked_retention(self):
    bucket_uri = self.CreateBucketWithRetentionPolicy(
        retention_period_in_seconds=_SECONDS_IN_DAY)
    self.RunGsUtil(['retention', 'clear', suri(bucket_uri)])
    self.VerifyRetentionPolicy(bucket_uri,
                               expected_retention_period_in_seconds=None)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_set_retention_unlocked_invalid_arg(self):
    bucket_uri = self.CreateBucket()
    stderr = self.RunGsUtil(
        ['retention', 'set', '1a', suri(bucket_uri)],
        expected_status=1,
        return_stderr=True)
    self.assertRegexpMatches(stderr, r'Incorrect retention period specified')

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_lock_retention_userConfirms(self):
    bucket_uri = self.CreateBucketWithRetentionPolicy(
        retention_period_in_seconds=_SECONDS_IN_DAY)
    self.RunGsUtil(['retention', 'lock', suri(bucket_uri)], stdin='y')
    self.VerifyRetentionPolicy(
        bucket_uri,
        expected_retention_period_in_seconds=_SECONDS_IN_DAY,
        expected_is_locked=True)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_lock_retention_userDoesNotConfirm(self):
    bucket_uri = self.CreateBucketWithRetentionPolicy(
        retention_period_in_seconds=_SECONDS_IN_DAY)
    stderr = self.RunGsUtil(
        ['retention', 'lock', suri(bucket_uri)], stdin='n', return_stderr=True)
    self.assertRegexpMatches(stderr, 'Abort Locking Retention Policy on')
    self.VerifyRetentionPolicy(
        bucket_uri, expected_retention_period_in_seconds=_SECONDS_IN_DAY)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_lock_with_no_retention_policy_invalid(self):
    bucket_uri = self.CreateBucket()
    stderr = self.RunGsUtil(
        ['retention', 'lock', suri(bucket_uri)],
        stdin='y',
        expected_status=1,
        return_stderr=True)
    self.assertRegexpMatches(stderr,
                             'does not have an Unlocked Retention Policy')
    self.VerifyRetentionPolicy(bucket_uri,
                               expected_retention_period_in_seconds=None)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_lock_retention_with_invalid_arg(self):
    bucket_uri = self.CreateBucket()
    stderr = self.RunGsUtil(
        ['retention', 'lock', '-a', suri(bucket_uri)],
        stdin='y',
        expected_status=1,
        return_stderr=True)
    self.assertRegexpMatches(stderr, r'Incorrect option\(s\) specified')

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_lock_retention_already_locked(self):
    bucket_uri = self.CreateBucketWithRetentionPolicy(
        retention_period_in_seconds=_SECONDS_IN_DAY, is_locked=True)
    stderr = self.RunGsUtil(
        ['retention', 'lock', suri(bucket_uri)], stdin='y', return_stderr=True)
    self.assertRegexpMatches(stderr,
                             r'Retention Policy on .* is already locked')

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_increase_retention_locked(self):
    bucket_uri = self.CreateBucketWithRetentionPolicy(
        retention_period_in_seconds=_SECONDS_IN_DAY, is_locked=True)
    self.RunGsUtil([
        'retention', 'set', '{}s'.format(_SECONDS_IN_DAY + 1),
        suri(bucket_uri)
    ])
    self.VerifyRetentionPolicy(
        bucket_uri,
        expected_retention_period_in_seconds=_SECONDS_IN_DAY + 1,
        expected_is_locked=True)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_decrease_retention_locked(self):
    bucket_uri = self.CreateBucketWithRetentionPolicy(
        retention_period_in_seconds=_SECONDS_IN_DAY, is_locked=True)
    stderr = self.RunGsUtil([
        'retention', 'set', '{}s'.format(_SECONDS_IN_DAY - 1),
        suri(bucket_uri)
    ],
                            expected_status=1,
                            return_stderr=True)
    self.assertRegexpMatches(
        stderr, '403 Cannot reduce retention duration of a '
        'locked Retention Policy for bucket')

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_clear_locked_retention(self):
    bucket_uri = self.CreateBucketWithRetentionPolicy(
        retention_period_in_seconds=_SECONDS_IN_DAY, is_locked=True)
    stderr = self.RunGsUtil(
        ['retention', 'clear', suri(bucket_uri)],
        expected_status=1,
        return_stderr=True)
    self.assertRegexpMatches(
        stderr,
        r'403 Bucket .* has a locked Retention Policy which cannot be removed')

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_get_retention_locked(self):
    bucket_uri = self.CreateBucketWithRetentionPolicy(
        retention_period_in_seconds=_SECONDS_IN_DAY, is_locked=True)
    stdout = self.RunGsUtil(
        ['retention', 'get', suri(bucket_uri)], return_stdout=True)
    self.assertRegexpMatches(stdout, r'Retention Policy \(LOCKED\):')
    self.assertRegexpMatches(stdout, r'Duration: 1 Day\(s\)')
    self.assertRegexpMatches(stdout, r'Effective Time: .* GMT')
    expected_effective_time = self._ConvertTimeStringToSeconds(
        re.search(r'(?<=Time: )[\w,: ]+', stdout).group())
    actual_retention_policy = self.json_api.GetBucket(
        bucket_uri.bucket_name, fields=['retentionPolicy']).retentionPolicy
    actual_effective_time = self.DateTimeToSeconds(
        actual_retention_policy.effectiveTime.replace(tzinfo=None))
    self.assertEqual(actual_effective_time, expected_effective_time)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_get_retention_unlocked(self):
    bucket_uri = self.CreateBucketWithRetentionPolicy(
        retention_period_in_seconds=_SECONDS_IN_DAY)
    stdout = self.RunGsUtil(
        ['retention', 'get', suri(bucket_uri)], return_stdout=True)
    self.assertRegexpMatches(stdout, r'Retention Policy \(UNLOCKED\):')
    self.assertRegexpMatches(stdout, r'Duration: 1 Day\(s\)')
    self.assertRegexpMatches(stdout, r'Effective Time: .* GMT')
    expected_effective_time = self._ConvertTimeStringToSeconds(
        re.search(r'(?<=Time: )[\w,: ]+', stdout).group())
    actual_retention_policy = self.json_api.GetBucket(
        bucket_uri.bucket_name, fields=['retentionPolicy']).retentionPolicy
    actual_effective_time = self.DateTimeToSeconds(
        actual_retention_policy.effectiveTime.replace(tzinfo=None))
    self.assertEqual(actual_effective_time, expected_effective_time)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_get_no_retention(self):
    bucket_uri = self.CreateBucket()
    stdout = self.RunGsUtil(
        ['retention', 'get', suri(bucket_uri)], return_stdout=True)
    self.assertRegexpMatches(stdout, 'has no Retention Policy')

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_get_invalid_args(self):
    bucket_uri = self.CreateBucketWithRetentionPolicy(
        retention_period_in_seconds=_SECONDS_IN_DAY)
    stderr = self.RunGsUtil(
        ['retention', 'get', '-a', suri(bucket_uri)],
        expected_status=1,
        return_stderr=True)
    self.assertRegexpMatches(stderr, r'Incorrect option\(s\) specified.')

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_set_temporary_hold_invalid_arg(self):
    object_uri = self.CreateObject()
    stderr = self.RunGsUtil(['retention', 'temp', 'held',
                             suri(object_uri)],
                            expected_status=1,
                            return_stderr=True)
    self.assertRegexpMatches(
        stderr, r'Invalid subcommand ".*" for the "retention temp" command')

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_temporary_hold_bucket_with_no_retention(self):
    bucket_uri = self.CreateBucket()
    object_uri = self.CreateObject(bucket_uri=bucket_uri, contents='content')
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object_uri,
                                             temporary_hold=None)
    self.RunGsUtil(['retention', 'temp', 'set', suri(object_uri)])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object_uri,
                                             temporary_hold=True)
    self.RunGsUtil(['retention', 'temp', 'release', suri(object_uri)])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object_uri,
                                             temporary_hold=False)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_temporary_hold_bucket_with_retention(self):
    retention_period = 1
    bucket_uri = self.CreateBucketWithRetentionPolicy(retention_period)
    object_uri = self.CreateObject(bucket_uri=bucket_uri, contents='content')
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object_uri,
                                             temporary_hold=None,
                                             retention_period=retention_period)
    self.RunGsUtil(['retention', 'temp', 'set', suri(object_uri)])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object_uri,
                                             temporary_hold=True,
                                             retention_period=retention_period)
    self.RunGsUtil(['retention', 'temp', 'release', suri(object_uri)])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object_uri,
                                             temporary_hold=False,
                                             retention_period=retention_period)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_temporary_hold_multiple_sequential(self):
    bucket_uri = self.CreateBucket()
    object1_uri = self.CreateObject(bucket_uri=bucket_uri, contents='content')
    object2_uri = self.CreateObject(bucket_uri=bucket_uri, contents='content')
    self.RunGsUtil(
        ['retention', 'temp', 'set',
         suri(object1_uri),
         suri(object2_uri)])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object1_uri,
                                             temporary_hold=True)
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object2_uri,
                                             temporary_hold=True)
    self.RunGsUtil(
        ['retention', 'temp', 'release',
         suri(object1_uri),
         suri(object2_uri)])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object1_uri,
                                             temporary_hold=False)
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object2_uri,
                                             temporary_hold=False)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_temporary_hold_multiple_parallel(self):
    bucket_uri = self.CreateBucket()
    object1_uri = self.CreateObject(bucket_uri=bucket_uri, contents='content')
    object2_uri = self.CreateObject(bucket_uri=bucket_uri, contents='content')
    self.RunGsUtil([
        '-m', 'retention', 'temp', 'set',
        suri(object1_uri),
        suri(object2_uri)
    ])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object1_uri,
                                             temporary_hold=True)
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object2_uri,
                                             temporary_hold=True)
    self.RunGsUtil([
        '-m', 'retention', 'temp', 'release',
        suri(object1_uri),
        suri(object2_uri)
    ])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object1_uri,
                                             temporary_hold=False)
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object2_uri,
                                             temporary_hold=False)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_set_event_based_hold_invalid_arg(self):
    object_uri = self.CreateObject()
    stderr = self.RunGsUtil(['retention', 'event', 'rel',
                             suri(object_uri)],
                            expected_status=1,
                            return_stderr=True)
    self.assertRegexpMatches(
        stderr, r'Invalid subcommand ".*" for the "retention event" command')

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_event_based_hold_bucket_with_no_retention(self):
    bucket_uri = self.CreateBucket()
    object_uri = self.CreateObject(bucket_uri=bucket_uri, contents='content')
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object_uri,
                                             event_based_hold=None)
    self.RunGsUtil(['retention', 'event', 'set', suri(object_uri)])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object_uri,
                                             event_based_hold=True)
    self.RunGsUtil(['retention', 'event', 'release', suri(object_uri)])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object_uri,
                                             event_based_hold=False)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_event_based_hold_bucket_with_retention(self):
    retention_period = 1
    bucket_uri = self.CreateBucketWithRetentionPolicy(retention_period)
    object_uri = self.CreateObject(bucket_uri=bucket_uri, contents='content')
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object_uri,
                                             event_based_hold=None,
                                             retention_period=retention_period)
    self.RunGsUtil(['retention', 'event', 'set', suri(object_uri)])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object_uri,
                                             event_based_hold=True,
                                             retention_period=None)
    self.RunGsUtil(['retention', 'event', 'release', suri(object_uri)])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object_uri,
                                             event_based_hold=False,
                                             retention_period=retention_period)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_event_based_hold_multiple_sequential(self):
    bucket_uri = self.CreateBucket()
    object1_uri = self.CreateObject(bucket_uri=bucket_uri, contents='content')
    object2_uri = self.CreateObject(bucket_uri=bucket_uri, contents='content')
    self.RunGsUtil(
        ['retention', 'event', 'set',
         suri(object1_uri),
         suri(object2_uri)])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object1_uri,
                                             event_based_hold=True)
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object2_uri,
                                             event_based_hold=True)
    self.RunGsUtil(
        ['retention', 'event', 'release',
         suri(object1_uri),
         suri(object2_uri)])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object1_uri,
                                             event_based_hold=False)
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object2_uri,
                                             event_based_hold=False)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_event_based_hold_multiple_parallel(self):
    bucket_uri = self.CreateBucket()
    object1_uri = self.CreateObject(bucket_uri=bucket_uri, contents='content')
    object2_uri = self.CreateObject(bucket_uri=bucket_uri, contents='content')
    self.RunGsUtil([
        '-m', 'retention', 'event', 'set',
        suri(object1_uri),
        suri(object2_uri)
    ])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object1_uri,
                                             event_based_hold=True)
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object2_uri,
                                             event_based_hold=True)
    self.RunGsUtil([
        '-m', 'retention', 'event', 'release',
        suri(object1_uri),
        suri(object2_uri)
    ])
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object1_uri,
                                             event_based_hold=False)
    self._VerifyObjectHoldAndRetentionStatus(bucket_uri,
                                             object2_uri,
                                             event_based_hold=False)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_default_event_based_hold(self):
    bucket_uri = self.CreateBucket()
    self.RunGsUtil(['retention', 'event-default', 'set', suri(bucket_uri)])
    self._VerifyDefaultEventBasedHold(bucket_uri,
                                      expected_default_event_based_hold=True)
    self.RunGsUtil(['retention', 'event-default', 'release', suri(bucket_uri)])
    self._VerifyDefaultEventBasedHold(bucket_uri,
                                      expected_default_event_based_hold=False)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_default_event_based_hold_multiple_sequential(self):
    bucket1_uri = self.CreateBucket()
    bucket2_uri = self.CreateBucket()
    self.RunGsUtil([
        'retention', 'event-default', 'set',
        suri(bucket1_uri),
        suri(bucket2_uri)
    ])
    self._VerifyDefaultEventBasedHold(bucket1_uri,
                                      expected_default_event_based_hold=True)
    self._VerifyDefaultEventBasedHold(bucket2_uri,
                                      expected_default_event_based_hold=True)

  @SkipForS3('Retention is not supported for s3 objects')
  @SkipForXML('Retention is not supported for XML API')
  def test_default_event_based_hold_multiple_parallel(self):
    bucket1_uri = self.CreateBucket()
    bucket2_uri = self.CreateBucket()
    self.RunGsUtil([
        'retention', 'event-default', 'set',
        suri(bucket1_uri),
        suri(bucket2_uri)
    ])
    self._VerifyDefaultEventBasedHold(bucket1_uri,
                                      expected_default_event_based_hold=True)
    self._VerifyDefaultEventBasedHold(bucket2_uri,
                                      expected_default_event_based_hold=True)

  def _VerifyObjectHoldAndRetentionStatus(self,
                                          bucket_uri,
                                          object_uri,
                                          temporary_hold=None,
                                          event_based_hold=None,
                                          retention_period=None):
    object_metadata = self.json_api.GetObjectMetadata(
        bucket_uri.bucket_name,
        object_uri.object_name,
        fields=[
            'timeCreated', 'temporaryHold', 'eventBasedHold',
            'retentionExpirationTime'
        ])
    if temporary_hold is None:
      self.assertEqual(object_metadata.temporaryHold, None)
    else:
      self.assertEqual(object_metadata.temporaryHold, temporary_hold)

    if event_based_hold is None:
      self.assertEqual(object_metadata.eventBasedHold, None)
    else:
      self.assertEqual(object_metadata.eventBasedHold, event_based_hold)

    if retention_period is None:
      self.assertEqual(object_metadata.retentionExpirationTime, None)
    elif event_based_hold is False:
      retention_policy = self.json_api.GetBucket(bucket_uri.bucket_name,
                                                 fields=['retentionPolicy'
                                                        ]).retentionPolicy
      time_delta = datetime.timedelta(0, retention_policy.retentionPeriod)
      expected_expiration_time = object_metadata.timeCreated + time_delta
      if event_based_hold is None:
        self.assertEqual(object_metadata.retentionExpirationTime,
                         expected_expiration_time)
      else:
        # since we don't expose the release time of event-based hold we can
        # only verify that expected_expiration_time is greater than
        #     object-creation-time + retention period
        # that is because
        #     eventBased-hold's release time > object-creation-time
        self.assertGreater(object_metadata.retentionExpirationTime,
                           expected_expiration_time)

  def _VerifyDefaultEventBasedHold(self,
                                   bucket_uri,
                                   expected_default_event_based_hold=None):
    actual_default_event_based_hold = self.json_api.GetBucket(
        bucket_uri.bucket_name,
        fields=['defaultEventBasedHold']).defaultEventBasedHold

    if expected_default_event_based_hold is None:
      self.assertEqual(actual_default_event_based_hold, None)
    else:
      self.assertEqual(actual_default_event_based_hold,
                       expected_default_event_based_hold)

  def _ConvertTimeStringToSeconds(self, time_string):
    """Converts time in following format to its equivalent timestamp in seconds.

      Format: '%a, %d %b %Y %H:%M:%S GMT'
        i.e.: 'Fri, 18 Aug 2017 23:31:39 GMT'

    Args:
      time_string: time in string format.

    Returns:
      returns equivalent timestamp in seconds of given time.
    """
    converted_time = datetime.datetime.strptime(time_string,
                                                '%a, %d %b %Y %H:%M:%S GMT')
    return self.DateTimeToSeconds(converted_time)



# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Mock logging handler to check for expected logs."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import logging


class MockLoggingHandler(logging.Handler):
  """Mock logging handler to check for expected logs."""

  def __init__(self, *args, **kwargs):
    self.reset()
    logging.Handler.__init__(self, *args, **kwargs)

  def emit(self, record):
    self.messages[record.levelname.lower()].append(record.getMessage())

  def reset(self):
    self.messages = {
        'debug': [],
        'info': [],
        'warning': [],
        'error': [],
        'critical': [],
    }



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Setup installation module for gsutil."""

import os

from setuptools import find_packages
from setuptools import setup
from setuptools.command import build_py
from setuptools.command import sdist

long_desc_content_type = 'text/plain'

long_desc = """
gsutil is a Python application that lets you access Google Cloud Storage from
the command line. You can use gsutil to do a wide range of bucket and object
management tasks, including:
 * Creating and deleting buckets.
 * Uploading, downloading, and deleting objects.
 * Listing buckets and objects.
 * Moving, copying, and renaming objects.
 * Editing object and bucket ACLs.
"""

requires = [
    'argcomplete>=1.9.4',
    'crcmod>=1.7',
    'fasteners>=0.14.1',
    'gcs-oauth2-boto-plugin>=2.5',
    'google-apitools>=0.5.30',
    'httplib2>=0.11.3',
    'google-reauth>=0.1.0',
    # TODO: Sync submodule with tag referenced here once #339 is fixed in mock.
    'mock==2.0.0',
    'monotonic>=1.4',
    'oauth2client==4.1.3',
    'pyOpenSSL>=0.13',
    'retry_decorator>=1.0.0',
    'six>=1.12.0',
    # Not using 1.02 because of:
    #   https://code.google.com/p/socksipy-branch/issues/detail?id=3
    'SocksiPy-branch==1.01',
]

CURDIR = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(CURDIR, 'VERSION'), 'r') as f:
  VERSION = f.read().strip()

with open(os.path.join(CURDIR, 'CHECKSUM'), 'r') as f:
  CHECKSUM = f.read()


def PlaceNeededFiles(self, target_dir):
  """Populates necessary files into the gslib module and unit test modules."""
  target_dir = os.path.join(target_dir, 'gslib')
  self.mkpath(target_dir)

  # Copy the gsutil root VERSION file into gslib module.
  with open(os.path.join(target_dir, 'VERSION'), 'w') as fp:
    fp.write(VERSION)

  # Copy the gsutil root CHECKSUM file into gslib module.
  with open(os.path.join(target_dir, 'CHECKSUM'), 'w') as fp:
    fp.write(CHECKSUM)


class CustomBuildPy(build_py.build_py):
  """Excludes update command from package-installed versions of gsutil."""

  def byte_compile(self, files):
    for filename in files:
      # Note: we exclude the update command here because binary distributions
      # (built via setup.py bdist command) don't abide by the MANIFEST file.
      # For source distributions (built via setup.py sdist), the update command
      # will be excluded by the MANIFEST file.
      if 'gslib/commands/update.py' in filename:
        os.unlink(filename)
    build_py.build_py.byte_compile(self, files)

  def run(self):
    if not self.dry_run:
      PlaceNeededFiles(self, self.build_lib)
      build_py.build_py.run(self)


class CustomSDist(sdist.sdist):

  def make_release_tree(self, base_dir, files):
    sdist.sdist.make_release_tree(self, base_dir, files)
    PlaceNeededFiles(self, base_dir)


setup(
    name='gsutil',
    version=VERSION,
    url='https://cloud.google.com/storage/docs/gsutil',
    download_url='https://cloud.google.com/storage/docs/gsutil_install',
    license='Apache 2.0',
    author='Google Inc.',
    author_email='gs-team@google.com',
    description=('A command line tool for interacting with cloud storage '
                 'services.'),
    long_description=long_desc,
    long_description_content_type=long_desc_content_type,
    zip_safe=True,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: System :: Filesystems',
        'Topic :: Utilities',
    ],
    # Gsutil supports Python 2.7, 3.5+
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, <4',
    platforms='any',
    packages=find_packages(
        exclude=[
            # Packages under third_party are installed separately as
            # dependencies (see the "requires" list above). Note that we do not
            # exclude vendored dependencies (under gslib/vendored), as our
            # vendored versions may differ slightly from the official versions.
            'third_party',
        ],
    ),
    include_package_data=True,
    entry_points={
        'console_scripts': ['gsutil = gslib.__main__:main',],
    },
    install_requires=requires,
    cmdclass={
        'build_py': CustomBuildPy,
        'sdist': CustomSDist,
    },
)



# -*- coding: utf-8 -*-
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for retention_util module."""

from __future__ import absolute_import

import gslib.tests.testcase as testcase
from gslib.utils.retention_util import _RetentionPeriodToString
from gslib.utils.retention_util import DaysToSeconds
from gslib.utils.retention_util import MonthsToSeconds
from gslib.utils.retention_util import RetentionInDaysMatch
from gslib.utils.retention_util import RetentionInMonthsMatch
from gslib.utils.retention_util import RetentionInSeconds
from gslib.utils.retention_util import RetentionInSecondsMatch
from gslib.utils.retention_util import RetentionInYearsMatch
from gslib.utils.retention_util import SECONDS_IN_DAY
from gslib.utils.retention_util import SECONDS_IN_MONTH
from gslib.utils.retention_util import SECONDS_IN_YEAR
from gslib.utils.retention_util import YearsToSeconds


class TestRetentionUtil(testcase.GsUtilUnitTestCase):
  """Unit tests for gsutil retention_util module."""

  def testDaysToSeconds(self):
    secs = DaysToSeconds(1)
    self.assertEqual(secs, 1 * SECONDS_IN_DAY)

    secs = DaysToSeconds(3)
    self.assertEqual(secs, 3 * SECONDS_IN_DAY)

  def testMonthsToSeconds(self):
    secs = MonthsToSeconds(1)
    self.assertEqual(secs, 1 * SECONDS_IN_MONTH)

    secs = MonthsToSeconds(3)
    self.assertEqual(secs, 3 * SECONDS_IN_MONTH)

  def testYearsToSeconds(self):
    secs = YearsToSeconds(1)
    self.assertEqual(secs, 1 * SECONDS_IN_YEAR)

    secs = YearsToSeconds(3)
    self.assertEqual(secs, 3 * SECONDS_IN_YEAR)

  def testRetentionInSecondsMatch(self):
    secs = '30s'
    secs_match = RetentionInSecondsMatch(secs)
    self.assertEquals('30', secs_match.group('number'))

    secs = '1s'
    secs_match = RetentionInSecondsMatch(secs)
    self.assertEquals('1', secs_match.group('number'))

    secs = '1second'
    secs_match = RetentionInSecondsMatch(secs)
    self.assertEquals(None, secs_match)

  def testRetentionInMonthsMatch(self):
    months = '30m'
    months_match = RetentionInMonthsMatch(months)
    self.assertEquals('30', months_match.group('number'))

    months = '1m'
    months_match = RetentionInMonthsMatch(months)
    self.assertEquals('1', months_match.group('number'))

    months = '1month'
    months_match = RetentionInMonthsMatch(months)
    self.assertEquals(None, months_match)

  def testRetentionInDaysMatch(self):
    days = '30d'
    days_match = RetentionInDaysMatch(days)
    self.assertEquals('30', days_match.group('number'))

    days = '1d'
    days_match = RetentionInDaysMatch(days)
    self.assertEquals('1', days_match.group('number'))

    days = '1day'
    days_match = RetentionInDaysMatch(days)
    self.assertEquals(None, days_match)

  def testRetentionInYearsMatch(self):
    years = '30y'
    years_match = RetentionInYearsMatch(years)
    self.assertEquals('30', years_match.group('number'))

    years = '1y'
    years_match = RetentionInYearsMatch(years)
    self.assertEquals('1', years_match.group('number'))

    years = '1year'
    years_match = RetentionInYearsMatch(years)
    self.assertEquals(None, years_match)

  def testRetentionInSeconds(self):
    one_year = '1y'
    one_year_in_seconds = RetentionInSeconds(one_year)
    self.assertEquals(SECONDS_IN_YEAR, one_year_in_seconds)

    one_month = '1m'
    one_month_in_seconds = RetentionInSeconds(one_month)
    self.assertEquals(SECONDS_IN_MONTH, one_month_in_seconds)

    one_day = '1d'
    one_day_in_seconds = RetentionInSeconds(one_day)
    self.assertEquals(SECONDS_IN_DAY, one_day_in_seconds)

    one_second = '1s'
    one_second_in_seconds = RetentionInSeconds(one_second)
    self.assertEquals(1, one_second_in_seconds)

  def testRetentionPeriodToString(self):
    retention_str = _RetentionPeriodToString(SECONDS_IN_DAY)
    self.assertRegexpMatches(retention_str, r'Duration: 1 Day\(s\)')

    retention_str = _RetentionPeriodToString(SECONDS_IN_DAY - 1)
    self.assertRegexpMatches(retention_str, r'Duration: 86399 Second\(s\)')

    retention_str = _RetentionPeriodToString(SECONDS_IN_DAY + 1)
    self.assertRegexpMatches(retention_str,
                             r'Duration: 86401 Seconds \(~1 Day\(s\)\)')

    retention_str = _RetentionPeriodToString(SECONDS_IN_MONTH)
    self.assertRegexpMatches(retention_str, r'Duration: 1 Month\(s\)')

    retention_str = _RetentionPeriodToString(SECONDS_IN_MONTH - 1)
    self.assertRegexpMatches(retention_str,
                             r'Duration: 2678399 Seconds \(~30 Day\(s\)\)')

    retention_str = _RetentionPeriodToString(SECONDS_IN_MONTH + 1)
    self.assertRegexpMatches(retention_str,
                             r'Duration: 2678401 Seconds \(~31 Day\(s\)\)')

    retention_str = _RetentionPeriodToString(SECONDS_IN_YEAR)
    self.assertRegexpMatches(retention_str, r'Duration: 1 Year\(s\)')

    retention_str = _RetentionPeriodToString(SECONDS_IN_YEAR - 1)
    self.assertRegexpMatches(retention_str,
                             r'Duration: 31557599 Seconds \(~365 Day\(s\)\)')

    retention_str = _RetentionPeriodToString(SECONDS_IN_YEAR + 1)
    self.assertRegexpMatches(retention_str,
                             r'Duration: 31557601 Seconds \(~365 Day\(s\)\)')



# -*- coding: utf-8 -*-
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for stat command."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from gslib.cs_api_map import ApiSelector
from gslib.exception import NO_URLS_MATCHED_TARGET
import gslib.tests.testcase as testcase
from gslib.tests.testcase.integration_testcase import SkipForS3
from gslib.tests.testcase.integration_testcase import SkipForXML
from gslib.tests.util import GenerationFromURI as urigen
from gslib.tests.util import ObjectToURI as suri
from gslib.tests.util import SetBotoConfigForTest
from gslib.tests.util import TEST_ENCRYPTION_CONTENT1
from gslib.tests.util import TEST_ENCRYPTION_CONTENT1_CRC32C
from gslib.tests.util import TEST_ENCRYPTION_CONTENT1_MD5
from gslib.tests.util import TEST_ENCRYPTION_CONTENT2
from gslib.tests.util import TEST_ENCRYPTION_CONTENT2_CRC32C
from gslib.tests.util import TEST_ENCRYPTION_CONTENT2_MD5
from gslib.tests.util import TEST_ENCRYPTION_CONTENT3
from gslib.tests.util import TEST_ENCRYPTION_CONTENT3_CRC32C
from gslib.tests.util import TEST_ENCRYPTION_CONTENT3_MD5
from gslib.tests.util import TEST_ENCRYPTION_KEY1
from gslib.tests.util import TEST_ENCRYPTION_KEY1_SHA256_B64
from gslib.tests.util import TEST_ENCRYPTION_KEY2
from gslib.tests.util import TEST_ENCRYPTION_KEY2_SHA256_B64
from gslib.tests.util import unittest
from gslib.utils.retry_util import Retry


class TestStat(testcase.GsUtilIntegrationTestCase):
  """Integration tests for stat command."""

  @SkipForS3('\'Archived time\' is a GS-specific response field.')
  @SkipForXML(
      'XML API only supports \'DeletedTime\' response field when making a '
      'GET Bucket request to list all objects, which is heavy overhead when '
      'the real intent is just a HEAD Object call.')
  def test_versioned_stat_output(self):
    """Tests stat output of an outdated object under version control."""
    bucket_uri = self.CreateVersionedBucket()
    old_object_uri = self.CreateObject(bucket_uri=bucket_uri, contents=b'z')

    # Update object
    self.CreateObject(bucket_uri=bucket_uri,
                      object_name=old_object_uri.object_name,
                      contents=b'z',
                      gs_idempotent_generation=urigen(old_object_uri))

    stdout = self.RunGsUtil(['stat', old_object_uri.version_specific_uri],
                            return_stdout=True)

    self.assertIn('Archived time', stdout)

  def test_stat_output(self):
    """Tests stat output of a single object."""
    object_uri = self.CreateObject(contents=b'z')
    stdout = self.RunGsUtil(['stat', suri(object_uri)], return_stdout=True)
    self.assertIn(object_uri.uri, stdout)
    self.assertIn('Creation time:', stdout)

    # Cache-Control and Content-Encoding can be different depending on
    # whether the JSON or XML API is used.  For JSON, only max-age and
    # no-cache are respected.  Although the object field will be populated
    # with whatever we set, the actual header returned from the JSON API
    # may differ from it (and differ from the XML response for the same object).
    #
    # Likewise, with contentEncoding, the field value and the header value
    # are not guaranteed to match or be the same across APIs.
    #
    # JSON will not return a Cache-control or content-encoding with the
    # current test object creation, so check these only for the XML API.
    if self.default_provider == 'gs':
      if self.test_api == ApiSelector.XML:
        self.assertIn('Cache-Control:', stdout)
        self.assertIn('Content-Encoding:', stdout)
      # TODO: Remove JSON check after adding storage class parsing in Boto.
      elif self.test_api == ApiSelector.JSON:
        self.assertIn('Storage class:', stdout)
      self.assertIn('Generation:', stdout)
      self.assertIn('Metageneration:', stdout)
      self.assertIn('Hash (crc32c):', stdout)
      self.assertIn('Hash (md5):', stdout)
      self.assertNotIn('Archived time', stdout)  # object is not archived
    self.assertIn('Content-Length:', stdout)
    self.assertIn('Content-Type:', stdout)
    self.assertIn('ETag:', stdout)

  def test_minus_q_stat(self):
    object_uri = self.CreateObject(contents=b'z')
    stdout = self.RunGsUtil(['-q', 'stat', suri(object_uri)],
                            return_stdout=True)
    self.assertEquals(0, len(stdout))
    stdout = self.RunGsUtil(
        ['-q', 'stat', suri(object_uri, 'junk')],
        return_stdout=True,
        expected_status=1)
    self.assertEquals(0, len(stdout))

  def test_stat_of_non_object_uri(self):
    self.RunGsUtil(['-q', 'stat', 'gs://'], expected_status=1)
    self.RunGsUtil(['-q', 'stat', 'gs://bucket/object'], expected_status=1)
    self.RunGsUtil(['-q', 'stat', 'file://tmp/abc'], expected_status=1)

  def test_stat_one_missing(self):
    bucket_uri = self.CreateBucket()
    self.CreateObject(bucket_uri=bucket_uri,
                      object_name='notmissing',
                      contents=b'z')
    stdout, stderr = self.RunGsUtil(
        ['stat',
         suri(bucket_uri, 'missing'),
         suri(bucket_uri, 'notmissing')],
        expected_status=1,
        return_stdout=True,
        return_stderr=True)
    self.assertIn(NO_URLS_MATCHED_TARGET % suri(bucket_uri, 'missing'), stderr)
    self.assertIn('%s:' % suri(bucket_uri, 'notmissing'), stdout)

  def test_stat_one_missing_wildcard(self):
    bucket_uri = self.CreateBucket()
    self.CreateObject(bucket_uri=bucket_uri,
                      object_name='notmissing',
                      contents=b'z')
    stdout, stderr = self.RunGsUtil(
        ['stat',
         suri(bucket_uri, 'missin*'),
         suri(bucket_uri, 'notmissin*')],
        expected_status=1,
        return_stdout=True,
        return_stderr=True)
    self.assertIn(NO_URLS_MATCHED_TARGET % suri(bucket_uri, 'missin*'), stderr)
    self.assertIn('%s:' % suri(bucket_uri, 'notmissing'), stdout)

  def test_stat_bucket_wildcard(self):
    # Prefix is a workaround for XML API limitation, see PR 766 for details
    bucket_uri = self.CreateBucket(bucket_name_prefix='aaa-')
    self.CreateObject(bucket_uri=bucket_uri, object_name='foo', contents=b'z')
    stat_string = suri(bucket_uri)[:-1] + '?/foo'
    self.RunGsUtil(['stat', stat_string])
    stat_string2 = suri(bucket_uri)[:-1] + '*/foo'
    self.RunGsUtil(['stat', stat_string2])

  def test_stat_object_wildcard(self):
    bucket_uri = self.CreateBucket()
    object1_uri = self.CreateObject(bucket_uri=bucket_uri,
                                    object_name='foo1',
                                    contents=b'z')
    object2_uri = self.CreateObject(bucket_uri=bucket_uri,
                                    object_name='foo2',
                                    contents=b'z')
    stat_string = suri(object1_uri)[:-2] + '*'

    # Use @Retry as hedge against bucket listing eventual consistency.
    @Retry(AssertionError, tries=3, timeout_secs=1)
    def _Check1():
      stdout = self.RunGsUtil(['stat', stat_string], return_stdout=True)
      self.assertIn(suri(object1_uri), stdout)
      self.assertIn(suri(object2_uri), stdout)

    _Check1()

  @SkipForS3('S3 customer-supplied encryption keys are not supported.')
  def test_stat_encrypted_object(self):
    """Tests stat command with an encrypted object."""
    if self.test_api == ApiSelector.XML:
      return unittest.skip(
          'gsutil does not support encryption with the XML API')
    bucket_uri = self.CreateBucket()
    object_uri = self.CreateObject(bucket_uri=bucket_uri,
                                   object_name='foo',
                                   contents=TEST_ENCRYPTION_CONTENT1,
                                   encryption_key=TEST_ENCRYPTION_KEY1)

    # Stat object with key should return unencrypted hashes.
    with SetBotoConfigForTest([('GSUtil', 'encryption_key',
                                TEST_ENCRYPTION_KEY1)]):
      stdout = self.RunGsUtil(['stat', suri(object_uri)], return_stdout=True)
      self.assertIn(TEST_ENCRYPTION_CONTENT1_MD5, stdout)
      self.assertIn(TEST_ENCRYPTION_CONTENT1_CRC32C, stdout)
      self.assertIn(TEST_ENCRYPTION_KEY1_SHA256_B64.decode('ascii'), stdout)

    # Stat object without key should return encrypted hashes.
    stdout = self.RunGsUtil(['stat', suri(object_uri)], return_stdout=True)
    self.assertNotIn(TEST_ENCRYPTION_CONTENT1_MD5, stdout)
    self.assertNotIn(TEST_ENCRYPTION_CONTENT1_CRC32C, stdout)
    self.assertIn('encrypted', stdout)
    self.assertIn(TEST_ENCRYPTION_KEY1_SHA256_B64.decode('ascii'), stdout)

  def test_stat_encrypted_object_wildcard(self):
    """Tests stat command with a mix of encrypted and unencrypted objects."""
    if self.test_api == ApiSelector.XML:
      return unittest.skip(
          'gsutil does not support encryption with the XML API')
    bucket_uri = self.CreateBucket()
    object1_uri = self.CreateObject(bucket_uri=bucket_uri,
                                    object_name='foo1',
                                    contents=TEST_ENCRYPTION_CONTENT1,
                                    encryption_key=TEST_ENCRYPTION_KEY1)
    object2_uri = self.CreateObject(bucket_uri=bucket_uri,
                                    object_name='foo2',
                                    contents=TEST_ENCRYPTION_CONTENT2,
                                    encryption_key=TEST_ENCRYPTION_KEY2)
    object3_uri = self.CreateObject(bucket_uri=bucket_uri,
                                    object_name='foo3',
                                    contents=TEST_ENCRYPTION_CONTENT3)

    stat_string = suri(object1_uri)[:-2] + '*'

    # Stat 3 objects, two encrypted each with a different key, and one
    # unencrypted. Should result in two unencrypted listing and one encrypted
    # listing.
    with SetBotoConfigForTest([('GSUtil', 'encryption_key',
                                TEST_ENCRYPTION_KEY1)]):
      # Use @Retry as hedge against bucket listing eventual consistency.
      @Retry(AssertionError, tries=3, timeout_secs=1)
      def _StatExpectMixed():
        """Runs stat and validates output."""
        stdout, stderr = self.RunGsUtil(['stat', stat_string],
                                        return_stdout=True,
                                        return_stderr=True)
        self.assertIn(suri(object1_uri), stdout)
        self.assertIn(TEST_ENCRYPTION_CONTENT1_MD5, stdout)
        self.assertIn(TEST_ENCRYPTION_CONTENT1_CRC32C, stdout)
        self.assertIn(TEST_ENCRYPTION_KEY1_SHA256_B64.decode('ascii'), stdout)
        self.assertIn(suri(object2_uri), stdout)
        self.assertNotIn(TEST_ENCRYPTION_CONTENT2_MD5, stdout)
        self.assertNotIn(TEST_ENCRYPTION_CONTENT2_CRC32C, stdout)
        self.assertIn('encrypted', stdout)
        self.assertIn(TEST_ENCRYPTION_KEY2_SHA256_B64.decode('ascii'), stdout)
        self.assertIn(suri(object3_uri), stdout)
        self.assertIn(TEST_ENCRYPTION_CONTENT3_MD5, stdout)
        self.assertIn(TEST_ENCRYPTION_CONTENT3_CRC32C, stdout)

      _StatExpectMixed()



# -*- coding: utf-8 -*-
# Copyright 2019 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of credentials that refreshes using the iamcredentials API."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import datetime

from oauth2client import client

from gslib.iamcredentials_api import IamcredentailsApi


class ImpersonationCredentials(client.OAuth2Credentials):
  _EXPIRY_FORMAT = '%Y-%m-%dT%H:%M:%SZ'

  def __init__(self, service_account_id, scopes, credentials, logger):
    self._service_account_id = service_account_id
    self.api = IamcredentailsApi(logger, credentials)

    response = self.api.GenerateAccessToken(service_account_id, scopes)
    self.access_token = response.accessToken
    self.token_expiry = self._ConvertExpiryTime(response.expireTime)

    super(ImpersonationCredentials, self).__init__(self.access_token,
                                                   None,
                                                   None,
                                                   None,
                                                   self.token_expiry,
                                                   None,
                                                   None,
                                                   scopes=scopes)

  @property
  def service_account_id(self):
    return self._service_account_id

  def _refresh(self, http):
    # client.Oauth2Credentials converts scopes into a set, so we need to convert
    # back to a list before making the API request.
    response = self.api.GenerateAccessToken(self._service_account_id,
                                            list(self.scopes))
    self.access_token = response.accessToken
    self.token_expiry = self._ConvertExpiryTime(response.expireTime)

  def _ConvertExpiryTime(self, value):
    return datetime.datetime.strptime(value,
                                      ImpersonationCredentials._EXPIRY_FORMAT)



# -*- coding: utf-8 -*-
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper functions for dealing with encryption keys used with cloud APIs."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import base64
import binascii
from hashlib import sha256
import re
import sys

import six
from gslib.exception import CommandException
from gslib.lazy_wrapper import LazyWrapper

MAX_DECRYPTION_KEYS = 100
VALID_CMEK_RE = LazyWrapper(
    lambda: re.compile('projects/([^/]+)/'
                       'locations/([a-zA-Z0-9_-]{1,63})/'
                       'keyRings/([a-zA-Z0-9_-]{1,63})/'
                       'cryptoKeys/([a-zA-Z0-9_-]{1,63})$'))


class CryptoKeyType(object):
  """Enum of valid types of encryption keys used with cloud API requests."""
  CSEK = 'CSEK'
  CMEK = 'CMEK'


class CryptoKeyWrapper(object):
  """Class describing a crypto key used with cloud API requests.

  This class should be instantiated via the `CryptoKeyWrapperFromKey` method.
  """

  def __init__(self, crypto_key):
    """Initialize the CryptoKeyWrapper.

    Args:
      crypto_key: Base64-encoded string of a CSEK, or the name of a Cloud KMS
          CMEK.

    Raises:
      CommandException: The specified crypto key was neither a CMEK key name nor
          a valid base64-encoded string of a CSEK.
    """
    self.crypto_key = crypto_key
    # Base64-encoded CSEKs always have a length of 44 characters, whereas
    # fully-qualified CMEK names are guaranteed to be longer than 45 characters.
    if len(crypto_key) == 44:
      self.crypto_type = CryptoKeyType.CSEK
      self.crypto_alg = 'AES256'  # Only currently supported algorithm for CSEK.
      try:
        self.crypto_key_sha256 = Base64Sha256FromBase64EncryptionKey(crypto_key)
      except:
        raise CommandException(
            'Configured encryption_key or decryption_key looked like a CSEK, '
            'but it was not a valid 44-character base64 string. Please '
            'double-check your configuration and ensure the key is correct.')
    else:  # CMEK
      try:
        ValidateCMEK(crypto_key)
      except CommandException as e:
        raise CommandException(
            'Configured encryption_key or decryption_key looked like a CMEK, '
            'but the key failed validation:\n%s' % e.reason)
      self.crypto_type = CryptoKeyType.CMEK
      self.crypto_alg = None
      self.crypto_key_sha256 = None


def CryptoKeyWrapperFromKey(crypto_key):
  """Returns a CryptoKeyWrapper for crypto_key, or None for no key."""
  return CryptoKeyWrapper(crypto_key) if crypto_key else None


def FindMatchingCSEKInBotoConfig(key_sha256, boto_config):
  """Searches boto_config for a CSEK with the given base64-encoded SHA256 hash.

  Args:
    key_sha256: (str) Base64-encoded SHA256 hash of the AES256 encryption key.
    boto_config: (boto.pyami.config.Config) The boto config in which to check
        for a matching encryption key.

  Returns:
    (str) Base64-encoded encryption key string if a match is found, None
    otherwise.
  """
  if six.PY3:
    if not isinstance(key_sha256, bytes):
      key_sha256 = key_sha256.encode('ascii')
  keywrapper = CryptoKeyWrapperFromKey(
      boto_config.get('GSUtil', 'encryption_key', None))
  if (keywrapper is not None and
      keywrapper.crypto_type == CryptoKeyType.CSEK and
      keywrapper.crypto_key_sha256 == key_sha256):
    return keywrapper.crypto_key

  for i in range(MAX_DECRYPTION_KEYS):
    key_number = i + 1
    keywrapper = CryptoKeyWrapperFromKey(
        boto_config.get('GSUtil', 'decryption_key%s' % str(key_number), None))
    if keywrapper is None:
      # Reading 100 config values can take ~1ms in testing. To avoid adding
      # this tax, stop reading keys as soon as we encounter a non-existent
      # entry (in lexicographic order).
      break
    elif (keywrapper.crypto_type == CryptoKeyType.CSEK and
          keywrapper.crypto_key_sha256 == key_sha256):
      return keywrapper.crypto_key


def GetEncryptionKeyWrapper(boto_config):
  """Returns a CryptoKeyWrapper for the configured encryption key.

  Reads in the value of the "encryption_key" attribute in boto_config, and if
  present, verifies it is a valid base64-encoded string and returns a
  CryptoKeyWrapper for it.

  Args:
    boto_config: (boto.pyami.config.Config) The boto config in which to check
        for a matching encryption key.

  Returns:
    CryptoKeyWrapper for the specified encryption key, or None if no encryption
    key was specified in boto_config.
  """
  encryption_key = boto_config.get('GSUtil', 'encryption_key', None)
  return CryptoKeyWrapper(encryption_key) if encryption_key else None


def Base64Sha256FromBase64EncryptionKey(csek_encryption_key):
  if six.PY3:
    if not isinstance(csek_encryption_key, bytes):
      csek_encryption_key = csek_encryption_key.encode('ascii')
  decoded_bytes = base64.decodestring(csek_encryption_key)
  key_sha256 = _CalculateSha256FromString(decoded_bytes)
  sha256_bytes = binascii.unhexlify(key_sha256)
  sha256_base64 = base64.encodestring(sha256_bytes)
  return sha256_base64.replace(b'\n', b'')


def ValidateCMEK(key):
  if not key:
    raise CommandException('KMS key is empty.')

  if key.startswith('/'):
    raise CommandException(
        'KMS key should not start with leading slash (/): "%s"' % key)

  if not VALID_CMEK_RE().match(key):
    raise CommandException(
        'Invalid KMS key name: "%s".\nKMS keys should follow the format '
        '"projects/<project-id>/locations/<location>/keyRings/<keyring>/'
        'cryptoKeys/<key-name>"' % key)


def _CalculateSha256FromString(input_string):
  sha256_hash = sha256()
  sha256_hash.update(input_string)
  return sha256_hash.hexdigest()


def _GetAndVerifyBase64EncryptionKey(boto_config):
  """Reads the encryption key from boto_config and ensures it is base64-encoded.

  Args:
    boto_config: (boto.pyami.config.Config) The boto config in which to check
        for a matching encryption key.

  Returns:
    (str) Base64-encoded encryption key string, or None if no encryption key
    exists in configuration.

  """
  encryption_key = boto_config.get('GSUtil', 'encryption_key', None)
  if encryption_key:
    # Ensure the key has a valid encoding.
    try:
      base64.decodestring(encryption_key)
    except:
      raise CommandException(
          'Configured encryption_key is not a valid base64 string. Please '
          'double-check your configuration and ensure the key is valid and in '
          'base64 format.')
  return encryption_key



# -*- coding: utf-8 -*-
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Shim for backwards-compatibility for moving GCE credentials.

oauth2client loads credentials classes based on the module name where
they were created; this means that moving GceAssertionCredentials from
here to third_party requires a shim mapping the old name to the new
one. Once loaded, the credential will be re-serialized with the new
path, meaning that we can (at some point) consider removing this file.
"""

# TODO: Remove this module once this change has been around long
# enough that old credentials are likely to be rare.

from apitools.base.py import GceAssertionCredentials



# -*- coding: utf-8 -*-
# Copyright 2013 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Gsutil API delegator for interacting with cloud storage providers."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

import boto
from boto import config
from gslib.cloud_api import ArgumentException
from gslib.cloud_api import CloudApi
from gslib.cs_api_map import ApiMapConstants
from gslib.cs_api_map import ApiSelector
from gslib.exception import CommandException


class CloudApiDelegator(CloudApi):
  """Class that handles delegating requests to gsutil Cloud API implementations.

  This class is responsible for determining at runtime which gsutil Cloud API
  implementation should service the request based on the Cloud storage provider,
  command-level API support, and configuration file override.

  During initialization it takes as an argument a gsutil_api_map which maps
  providers to their default and supported gsutil Cloud API implementations
  (see comments in cs_api_map for details).

  Instantiation of multiple delegators per-thread is required for multiprocess
  and/or multithreaded operations. Calling methods on the same delegator in
  multiple threads is unsafe.
  """

  def __init__(self,
               bucket_storage_uri_class,
               gsutil_api_map,
               logger,
               status_queue,
               provider=None,
               debug=0,
               trace_token=None,
               perf_trace_token=None,
               user_project=None):
    """Performs necessary setup for delegating cloud storage requests.

    This function has different arguments than the gsutil Cloud API __init__
    function because of the delegation responsibilties of this class.

    Args:
      bucket_storage_uri_class: boto storage_uri class, used by APIs that
                                provide boto translation or mocking.
      gsutil_api_map: Map of providers and API selector tuples to api classes
                      which can be used to communicate with those providers.
      logger: logging.logger for outputting log messages.
      status_queue: Queue for relaying status to UI.
      provider: Default provider prefix describing cloud storage provider to
                connect to.
      debug: Debug level for the API implementation (0..3).
      trace_token: Apiary trace token to pass to API.
      perf_trace_token: Performance trace token to use when making API calls.
      user_project: Project to be billed for this project.
    """
    super(CloudApiDelegator, self).__init__(bucket_storage_uri_class,
                                            logger,
                                            status_queue,
                                            provider=provider,
                                            debug=debug,
                                            trace_token=trace_token,
                                            perf_trace_token=perf_trace_token,
                                            user_project=user_project)
    self.api_map = gsutil_api_map
    self.prefer_api = boto.config.get('GSUtil', 'prefer_api', '').upper()
    self.loaded_apis = {}

    if not self.api_map[ApiMapConstants.API_MAP]:
      raise ArgumentException('No apiclass supplied for gsutil Cloud API map.')

  def _GetApi(self, provider):
    """Returns a valid CloudApi for use by the caller.

    This function lazy-loads connection and credentials using the API map
    and credential store provided during class initialization.

    Args:
      provider: Provider to load API for. If None, class-wide default is used.

    Raises:
      ArgumentException if there is no matching API available in the API map.

    Returns:
      Valid API instance that can be used to communicate with the Cloud
      Storage provider.
    """
    provider = provider or self.provider
    if not provider:
      raise ArgumentException('No provider selected for _GetApi')

    provider = str(provider)
    if provider not in self.loaded_apis:
      self.loaded_apis[provider] = {}

    api_selector = self.GetApiSelector(provider)
    if api_selector not in self.loaded_apis[provider]:
      # Need to load the API.
      self._LoadApi(provider, api_selector)
    return self.loaded_apis[provider][api_selector]

  def _LoadApi(self, provider, api_selector):
    """Loads a CloudApi into the loaded_apis map for this class.

    Args:
      provider: Provider to load the API for.
      api_selector: cs_api_map.ApiSelector defining the API type.
    """
    if provider not in self.api_map[ApiMapConstants.API_MAP]:
      raise ArgumentException(
          'gsutil Cloud API map contains no entry for provider %s.' % provider)
    if api_selector not in self.api_map[ApiMapConstants.API_MAP][provider]:
      raise ArgumentException(
          'gsutil Cloud API map does not support API %s for provider %s.' %
          (api_selector, provider))
    self.loaded_apis[provider][api_selector] = (
        self.api_map[ApiMapConstants.API_MAP][provider][api_selector](
            self.bucket_storage_uri_class,
            self.logger,
            self.status_queue,
            provider=provider,
            debug=self.debug,
            trace_token=self.trace_token,
            perf_trace_token=self.perf_trace_token,
            user_project=self.user_project))

  def GetApiSelector(self, provider=None):
    """Returns a cs_api_map.ApiSelector based on input and configuration.

    Args:
      provider: Provider to return the ApiSelector for.  If None, class-wide
                default is used.

    Returns:
      cs_api_map.ApiSelector that will be used for calls to the delegator
      for this provider.
    """
    selected_provider = provider or self.provider
    if not selected_provider:
      raise ArgumentException('No provider selected for CloudApi')

    if (selected_provider not in self.api_map[ApiMapConstants.DEFAULT_MAP] or
        self.api_map[ApiMapConstants.DEFAULT_MAP][selected_provider] not in
        self.api_map[ApiMapConstants.API_MAP][selected_provider]):
      raise ArgumentException('No default api available for provider %s' %
                              selected_provider)

    if selected_provider not in self.api_map[ApiMapConstants.SUPPORT_MAP]:
      raise ArgumentException('No supported apis available for provider %s' %
                              selected_provider)

    api = self.api_map[ApiMapConstants.DEFAULT_MAP][selected_provider]

    using_gs_hmac = (
        provider == 'gs' and
        not config.has_option('Credentials', 'gs_oauth2_refresh_token') and
        not (config.has_option('Credentials', 'gs_service_client_id') and
             config.has_option('Credentials', 'gs_service_key_file')) and
        (config.has_option('Credentials', 'gs_access_key_id') and
         config.has_option('Credentials', 'gs_secret_access_key')))

    configured_encryption = (provider == 'gs' and
                             (config.has_option('GSUtil', 'encryption_key') or
                              config.has_option('GSUtil', 'decryption_key1')))

    if using_gs_hmac and configured_encryption:
      raise CommandException(
          'gsutil does not support HMAC credentials with customer-supplied '
          'encryption keys (CSEK) or customer-managed KMS encryption keys '
          '(CMEK). Please generate and include non-HMAC credentials '
          'in your .boto configuration file, or to access public encrypted '
          'objects, remove your HMAC credentials.')
    # If we have only HMAC credentials for Google Cloud Storage, we must use
    # the XML API as the JSON API does not support HMAC.
    #
    # Technically if we have only HMAC credentials, we should still be able to
    # access public read resources via the JSON API, but the XML API can do
    # that just as well. It is better to use it than inspect the credentials on
    # every HTTP call.
    elif using_gs_hmac:
      api = ApiSelector.XML
    # CSEK and CMEK encryption keys are currently only supported in the
    # JSON API implementation (GcsJsonApi). We can't stop XML API users from
    # interacting with encrypted objects, since we don't know the object is
    # encrypted until after the API call is made, but if they specify
    # configuration values we will use JSON.
    elif configured_encryption:
      api = ApiSelector.JSON
    # Try to force the user's preference to a supported API.
    elif self.prefer_api in (
        self.api_map[ApiMapConstants.SUPPORT_MAP][selected_provider]):
      api = self.prefer_api
    return api

  # For function docstrings, see CloudApi class.
  def GetBucket(self, bucket_name, provider=None, fields=None):
    return self._GetApi(provider).GetBucket(bucket_name, fields=fields)

  def GetBucketIamPolicy(self, bucket_name, provider=None, fields=None):
    return self._GetApi(provider).GetBucketIamPolicy(bucket_name, fields=fields)

  def SetBucketIamPolicy(self, bucket_name, policy, provider=None):
    return self._GetApi(provider).SetBucketIamPolicy(bucket_name, policy)

  def ListBuckets(self, project_id=None, provider=None, fields=None):
    return self._GetApi(provider).ListBuckets(project_id=project_id,
                                              fields=fields)

  def PatchBucket(self,
                  bucket_name,
                  metadata,
                  canned_acl=None,
                  canned_def_acl=None,
                  preconditions=None,
                  provider=None,
                  fields=None):
    return self._GetApi(provider).PatchBucket(bucket_name,
                                              metadata,
                                              canned_acl=canned_acl,
                                              canned_def_acl=canned_def_acl,
                                              preconditions=preconditions,
                                              fields=fields)

  def LockRetentionPolicy(self, bucket_name, metageneration, provider=None):
    return self._GetApi(provider).LockRetentionPolicy(bucket_name,
                                                      metageneration,
                                                      provider=provider)

  def CreateBucket(self,
                   bucket_name,
                   project_id=None,
                   metadata=None,
                   provider=None,
                   fields=None):
    return self._GetApi(provider).CreateBucket(bucket_name,
                                               project_id=project_id,
                                               metadata=metadata,
                                               fields=fields)

  def DeleteBucket(self, bucket_name, preconditions=None, provider=None):
    return self._GetApi(provider).DeleteBucket(bucket_name,
                                               preconditions=preconditions)

  def GetObjectIamPolicy(self,
                         bucket_name,
                         object_name,
                         generation=None,
                         provider=None,
                         fields=None):
    return self._GetApi(provider).GetObjectIamPolicy(bucket_name,
                                                     object_name,
                                                     generation,
                                                     fields=fields)

  def SetObjectIamPolicy(self,
                         bucket_name,
                         object_name,
                         policy,
                         generation=None,
                         provider=None):
    return self._GetApi(provider).SetObjectIamPolicy(bucket_name, object_name,
                                                     policy, generation)

  def ListObjects(self,
                  bucket_name,
                  prefix=None,
                  delimiter=None,
                  all_versions=None,
                  provider=None,
                  fields=None):
    return self._GetApi(provider).ListObjects(bucket_name,
                                              prefix=prefix,
                                              delimiter=delimiter,
                                              all_versions=all_versions,
                                              fields=fields)

  def GetObjectMetadata(self,
                        bucket_name,
                        object_name,
                        generation=None,
                        provider=None,
                        fields=None):
    return self._GetApi(provider).GetObjectMetadata(bucket_name,
                                                    object_name,
                                                    generation=generation,
                                                    fields=fields)

  def PatchObjectMetadata(self,
                          bucket_name,
                          object_name,
                          metadata,
                          canned_acl=None,
                          generation=None,
                          preconditions=None,
                          provider=None,
                          fields=None):
    return self._GetApi(provider).PatchObjectMetadata(
        bucket_name,
        object_name,
        metadata,
        canned_acl=canned_acl,
        generation=generation,
        preconditions=preconditions,
        fields=fields)

  def GetObjectMedia(self,
                     bucket_name,
                     object_name,
                     download_stream,
                     provider=None,
                     generation=None,
                     object_size=None,
                     compressed_encoding=False,
                     download_strategy=CloudApi.DownloadStrategy.ONE_SHOT,
                     start_byte=0,
                     end_byte=None,
                     progress_callback=None,
                     serialization_data=None,
                     digesters=None,
                     decryption_tuple=None):
    return self._GetApi(provider).GetObjectMedia(
        bucket_name,
        object_name,
        download_stream,
        compressed_encoding=compressed_encoding,
        download_strategy=download_strategy,
        start_byte=start_byte,
        end_byte=end_byte,
        generation=generation,
        object_size=object_size,
        progress_callback=progress_callback,
        serialization_data=serialization_data,
        digesters=digesters,
        decryption_tuple=decryption_tuple)

  def UploadObject(self,
                   upload_stream,
                   object_metadata,
                   size=None,
                   canned_acl=None,
                   preconditions=None,
                   progress_callback=None,
                   encryption_tuple=None,
                   provider=None,
                   fields=None,
                   gzip_encoded=False):
    return self._GetApi(provider).UploadObject(
        upload_stream,
        object_metadata,
        size=size,
        canned_acl=canned_acl,
        preconditions=preconditions,
        progress_callback=progress_callback,
        encryption_tuple=encryption_tuple,
        fields=fields,
        gzip_encoded=gzip_encoded)

  def UploadObjectStreaming(self,
                            upload_stream,
                            object_metadata,
                            canned_acl=None,
                            preconditions=None,
                            progress_callback=None,
                            encryption_tuple=None,
                            provider=None,
                            fields=None,
                            gzip_encoded=False):
    return self._GetApi(provider).UploadObjectStreaming(
        upload_stream,
        object_metadata,
        canned_acl=canned_acl,
        preconditions=preconditions,
        progress_callback=progress_callback,
        encryption_tuple=encryption_tuple,
        fields=fields,
        gzip_encoded=gzip_encoded)

  def UploadObjectResumable(self,
                            upload_stream,
                            object_metadata,
                            canned_acl=None,
                            preconditions=None,
                            size=None,
                            serialization_data=None,
                            tracker_callback=None,
                            progress_callback=None,
                            encryption_tuple=None,
                            provider=None,
                            fields=None,
                            gzip_encoded=False):
    return self._GetApi(provider).UploadObjectResumable(
        upload_stream,
        object_metadata,
        canned_acl=canned_acl,
        preconditions=preconditions,
        size=size,
        serialization_data=serialization_data,
        tracker_callback=tracker_callback,
        progress_callback=progress_callback,
        encryption_tuple=encryption_tuple,
        fields=fields,
        gzip_encoded=gzip_encoded)

  def CopyObject(self,
                 src_obj_metadata,
                 dst_obj_metadata,
                 src_generation=None,
                 canned_acl=None,
                 preconditions=None,
                 progress_callback=None,
                 max_bytes_per_call=None,
                 encryption_tuple=None,
                 decryption_tuple=None,
                 provider=None,
                 fields=None):
    return self._GetApi(provider).CopyObject(
        src_obj_metadata,
        dst_obj_metadata,
        src_generation=src_generation,
        canned_acl=canned_acl,
        preconditions=preconditions,
        progress_callback=progress_callback,
        max_bytes_per_call=max_bytes_per_call,
        encryption_tuple=encryption_tuple,
        decryption_tuple=decryption_tuple,
        fields=fields)

  def ComposeObject(self,
                    src_objs_metadata,
                    dst_obj_metadata,
                    preconditions=None,
                    encryption_tuple=None,
                    provider=None,
                    fields=None):
    return self._GetApi(provider).ComposeObject(
        src_objs_metadata,
        dst_obj_metadata,
        preconditions=preconditions,
        encryption_tuple=encryption_tuple,
        fields=fields)

  def DeleteObject(self,
                   bucket_name,
                   object_name,
                   preconditions=None,
                   generation=None,
                   provider=None):
    return self._GetApi(provider).DeleteObject(bucket_name,
                                               object_name,
                                               preconditions=preconditions,
                                               generation=generation)

  def WatchBucket(self,
                  bucket_name,
                  address,
                  channel_id,
                  token=None,
                  provider=None,
                  fields=None):
    return self._GetApi(provider).WatchBucket(bucket_name,
                                              address,
                                              channel_id,
                                              token=token,
                                              fields=fields)

  def StopChannel(self, channel_id, resource_id, provider=None):
    return self._GetApi(provider).StopChannel(channel_id, resource_id)

  def ListChannels(self, bucket_name, provider=None):
    return self._GetApi(provider).ListChannels(bucket_name)

  def GetProjectServiceAccount(self, project_number, provider=None):
    return self._GetApi(provider).GetProjectServiceAccount(project_number)

  def CreateNotificationConfig(self,
                               bucket_name,
                               pubsub_topic,
                               payload_format,
                               event_types=None,
                               custom_attributes=None,
                               object_name_prefix=None,
                               provider=None):
    return self._GetApi(provider).CreateNotificationConfig(
        bucket_name, pubsub_topic, payload_format, event_types,
        custom_attributes, object_name_prefix)

  def DeleteNotificationConfig(self, bucket_name, notification, provider=None):
    return self._GetApi(provider).DeleteNotificationConfig(
        bucket_name, notification)

  def ListNotificationConfigs(self, bucket_name, provider=None):
    return self._GetApi(provider).ListNotificationConfigs(bucket_name)

  def ListBucketAccessControls(self, bucket_name, provider=None):
    return self._GetApi(provider).ListBucketAccessControls(bucket_name)

  def ListObjectAccessControls(self, bucket_name, object_name, provider=None):
    return self._GetApi(provider).ListObjectAccessControls(
        bucket_name, object_name)

  def CreateHmacKey(self, project_id, service_account_email, provider=None):
    return self._GetApi(provider).CreateHmacKey(project_id,
                                                service_account_email)

  def DeleteHmacKey(self, project_id, access_id, provider=None):
    return self._GetApi(provider).DeleteHmacKey(project_id, access_id)

  def GetHmacKey(self, project_id, access_id, provider=None):
    return self._GetApi(provider).GetHmacKey(project_id, access_id)

  def ListHmacKeys(self,
                   project_id,
                   service_account_email,
                   show_deleted_keys=False,
                   provider=None):
    return self._GetApi(provider).ListHmacKeys(project_id,
                                               service_account_email,
                                               show_deleted_keys)

  def SignUrl(self, provider, method, duration, path, logger, region,
              signed_headers, string_to_sign_debug):
    return self._GetApi(provider).SignUrl(
        method=method,
        duration=duration,
        path=path,
        logger=logger,
        region=region,
        signed_headers=signed_headers,
        string_to_sign_debug=string_to_sign_debug)

  def UpdateHmacKey(self, project_id, access_id, state, etag, provider=None):
    return self._GetApi(provider).UpdateHmacKey(project_id, access_id, state,
                                                etag)

  def XmlPassThroughGetAcl(self, storage_url, def_obj_acl=False, provider=None):
    """XML compatibility function for getting ACLs.

    Args:
      storage_url: StorageUrl object.
      def_obj_acl: If true, get the default object ACL on a bucket.
      provider: Cloud storage provider to connect to.  If not present,
                class-wide default is used.

    Raises:
      ArgumentException for errors during input validation.
      ServiceException for errors interacting with cloud storage providers.

    Returns:
      ACL XML for the resource specified by storage_url.
    """
    return self._GetApi(provider).XmlPassThroughGetAcl(storage_url,
                                                       def_obj_acl=def_obj_acl)

  def XmlPassThroughSetAcl(self,
                           acl_text,
                           storage_url,
                           canned=True,
                           def_obj_acl=False,
                           provider=None):
    """XML compatibility function for setting ACLs.

    Args:
      acl_text: XML ACL or canned ACL string.
      storage_url: StorageUrl object.
      canned: If true, acl_text is treated as a canned ACL string.
      def_obj_acl: If true, set the default object ACL on a bucket.
      provider: Cloud storage provider to connect to.  If not present,
                class-wide default is used.

    Raises:
      ArgumentException for errors during input validation.
      ServiceException for errors interacting with cloud storage providers.

    Returns:
      None.
    """
    self._GetApi(provider).XmlPassThroughSetAcl(acl_text,
                                                storage_url,
                                                canned=canned,
                                                def_obj_acl=def_obj_acl)

  def XmlPassThroughGetCors(self, storage_url, provider=None):
    """XML compatibility function for getting CORS configuration on a bucket.

    Args:
      storage_url: StorageUrl object.
      provider: Cloud storage provider to connect to.  If not present,
                class-wide default is used.

    Raises:
      ArgumentException for errors during input validation.
      ServiceException for errors interacting with cloud storage providers.

    Returns:
      CORS configuration XML for the bucket specified by storage_url.
    """
    return self._GetApi(provider).XmlPassThroughGetCors(storage_url)

  def XmlPassThroughSetCors(self, cors_text, storage_url, provider=None):
    """XML compatibility function for setting CORS configuration on a bucket.

    Args:
      cors_text: Raw CORS XML string.
      storage_url: StorageUrl object.
      provider: Cloud storage provider to connect to.  If not present,
                class-wide default is used.

    Raises:
      ArgumentException for errors during input validation.
      ServiceException for errors interacting with cloud storage providers.

    Returns:
      None.
    """
    self._GetApi(provider).XmlPassThroughSetCors(cors_text, storage_url)

  def XmlPassThroughGetLifecycle(self, storage_url, provider=None):
    """XML compatibility function for getting lifecycle config on a bucket.

    Args:
      storage_url: StorageUrl object.
      provider: Cloud storage provider to connect to.  If not present,
                class-wide default is used.

    Raises:
      ArgumentException for errors during input validation.
      ServiceException for errors interacting with cloud storage providers.

    Returns:
      Lifecycle configuration XML for the bucket specified by storage_url.
    """
    return self._GetApi(provider).XmlPassThroughGetLifecycle(storage_url)

  def XmlPassThroughSetLifecycle(self,
                                 lifecycle_text,
                                 storage_url,
                                 provider=None):
    """XML compatibility function for setting lifecycle config on a bucket.

    Args:
      lifecycle_text: Raw lifecycle configuration XML string.
      storage_url: StorageUrl object.
      provider: Cloud storage provider to connect to.  If not present,
                class-wide default is used.

    Raises:
      ArgumentException for errors during input validation.
      ServiceException for errors interacting with cloud storage providers.

    Returns:
      None.
    """
    self._GetApi(provider).XmlPassThroughSetLifecycle(lifecycle_text,
                                                      storage_url)

  def XmlPassThroughGetLogging(self, storage_url, provider=None):
    """XML compatibility function for getting logging configuration on a bucket.

    Args:
      storage_url: StorageUrl object.
      provider: Cloud storage provider to connect to.  If not present,
                class-wide default is used.

    Raises:
      ArgumentException for errors during input validation.
      ServiceException for errors interacting with cloud storage providers.

    Returns:
      Logging configuration XML for the bucket specified by storage_url.
    """
    return self._GetApi(provider).XmlPassThroughGetLogging(storage_url)

  def XmlPassThroughSetTagging(self, tags_text, storage_url, provider=None):
    """XML compatibility function for setting tagging configuration on a bucket.

    This passthrough provides support for setting a tagging configuration
    (equivalent to a label configuration) on a cloud bucket.

    Args:
      tags_text: Raw tagging configuration XML string.
      storage_url: StorageUrl object.
      provider: Cloud storage provider to connect to.  If not present,
                class-wide default is used.

    Raises:
      ArgumentException for errors during input validation.
      ServiceException for errors interacting with cloud storage providers.

    Returns:
      None.
    """
    return self._GetApi(provider).XmlPassThroughSetTagging(
        tags_text, storage_url)

  def XmlPassThroughGetTagging(self, storage_url, provider=None):
    """XML compatibility function for getting tagging configuration on a bucket.

    Args:
      storage_url: StorageUrl object.
      provider: Cloud storage provider to connect to.  If not present,
                class-wide default is used.

    Raises:
      ArgumentException for errors during input validation.
      ServiceException for errors interacting with cloud storage providers.

    Returns:
      Tagging configuration XML for the bucket specified by storage_url.
    """
    return self._GetApi(provider).XmlPassThroughGetTagging(storage_url)

  def XmlPassThroughGetWebsite(self, storage_url, provider=None):
    """XML compatibility function for getting website configuration on a bucket.

    Args:
      storage_url: StorageUrl object.
      provider: Cloud storage provider to connect to.  If not present,
                class-wide default is used.

    Raises:
      ArgumentException for errors during input validation.
      ServiceException for errors interacting with cloud storage providers.

    Returns:
      Website configuration XML for the bucket specified by storage_url.
    """
    return self._GetApi(provider).XmlPassThroughGetWebsite(storage_url)



# -*- coding: utf-8 -*-
# Copyright 2018 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provides a message queue that discards all messages."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals


class DiscardMessagesQueue(object):
  """Emulates a Cloud API status queue but drops all messages.

  This is useful when you want to perform some operations but not have the UI
  thread display information about those ops (e.g. running a test or fetching
  the public gsutil tarball object's metadata to perform a version check).
  """

  # pylint: disable=invalid-name, unused-argument
  def put(self, message=None, timeout=None):
    pass

  # pylint: enable=invalid-name, unused-argument



# -*- coding: utf-8 -*-
# Copyright 2014 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Additional help text about retry handling."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from gslib.help_provider import HelpProvider

_DETAILED_HELP_TEXT = ("""
<B>RETRY STRATEGY</B>
  There are a number of reasons that gsutil operations can fail; some are not
  retryable, and require that the user take some action, for example:

  - Invalid credentials
  - Network unreachable because of a proxy configuration problem
  - Access denied, because the bucket or object you are trying to use has an
    ACL that doesn't permit the action you're trying to perform.

  In other cases errors are retryable - transient network failures and HTTP 429
  and 5xx error codes. For these cases, gsutil will retry using a truncated
  binary exponential backoff strategy:

  - Wait a random period between [0..1] seconds and retry;
  - If that fails, wait a random period between [0..2] seconds and retry;
  - If that fails, wait a random period between [0..4] seconds and retry;
  - And so on, up to a configurable maximum number of retries (default = 23),
  with each retry period bounded by a configurable maximum period of time
  (default = 60 seconds).

  Thus, by default, gsutil will retry 23 times over 1+2+4+8+16+32+60... seconds
  for about 10 minutes. You can adjust the number of retries and maximum delay
  of any individual retry by editing the num_retries and max_retry_delay
  configuration variables in the "[Boto]" section of the .boto config file.
  Most users shouldn't need to change these values.

  For data transfers (the gsutil cp and rsync commands), gsutil provides
  additional retry functionality, in the form of resumable transfers.
  Essentially, a transfer that was interrupted because of a transient error
  can be restarted without starting over from scratch. For more details
  about this, see the "RESUMABLE TRANSFERS" section of "gsutil help cp".
""")


class CommandOptions(HelpProvider):
  """Additional help text about retry handling."""

  # Help specification. See help_provider.py for documentation.
  help_spec = HelpProvider.HelpSpec(
      help_name='retries',
      help_name_aliases=['retry', 'backoff', 'reliability'],
      help_type='additional_help',
      help_one_line_summary='Retry Handling Strategy',
      help_text=_DETAILED_HELP_TEXT,
      subcommand_help_text={},
  )



# -*- coding: utf-8 -*-
# Copyright 2011 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation of Unix-like mv command for cloud storage providers."""

from __future__ import absolute_import
from __future__ import print_function
from __future__ import division
from __future__ import unicode_literals

from gslib.command import Command
from gslib.command_argument import CommandArgument
from gslib.commands.cp import CP_SUB_ARGS
from gslib.cs_api_map import ApiSelector
from gslib.exception import CommandException
from gslib.storage_url import StorageUrlFromString
from gslib.utils.constants import NO_MAX

_SYNOPSIS = """
  gsutil mv [-p] src_url dst_url
  gsutil mv [-p] src_url... dst_url
  gsutil mv [-p] -I dst_url
"""

_DETAILED_HELP_TEXT = ("""
<B>SYNOPSIS</B>
""" + _SYNOPSIS + """


<B>DESCRIPTION</B>
  The gsutil mv command allows you to move data between your local file
  system and the cloud, move data within the cloud, and move data between
  cloud storage providers. For example, to move all objects from a
  bucket to a local directory you could use:

    gsutil mv gs://my_bucket/* dir

  Similarly, to move all objects from a local directory to a bucket you could
  use:

    gsutil mv ./dir gs://my_bucket


<B>RENAMING GROUPS OF OBJECTS</B>
  You can use the gsutil mv command to rename all objects with a given prefix
  to have a new prefix. This is accomplished by copying each object to a new
  object with the desired name and deleting the old one. For example, the
  command:

    gsutil mv gs://my_bucket/oldprefix gs://my_bucket/newprefix

  would rename all objects under gs://my_bucket/oldprefix to be under
  gs://my_bucket/newprefix, otherwise preserving the naming structure.

  If you do a rename as specified above and you want to preserve ACLs, you
  should use the -p option (see OPTIONS).

  Note that when using mv to rename groups of objects with a common prefix
  you cannot specify the source URL using wildcards. You need to spell out
  the complete name:

    gsutil mv gs://my_bucket/oldprefix gs://my_bucket/newprefix

  If you have a large number of files to move you might want to use the
  gsutil -m option, to perform a multi-threaded/multi-processing move:

    gsutil -m mv gs://my_bucket/oldprefix gs://my_bucket/newprefix


<B>NON-ATOMIC OPERATION</B>
  Unlike the case with many file systems, the gsutil mv command does not
  perform a single atomic operation. Rather, it performs a copy from source
  to destination followed by removing the source for each object.

  A consequence of this is that, in addition to normal network and operation
  charges, if you move a Nearline Storage, Coldline Storage, or Archive Storage
  object, deletion and data retrieval charges apply. See the `documentation
  <https://cloud.google.com/storage/pricing>`_ for pricing details.


<B>OPTIONS</B>
  All options that are available for the gsutil cp command are also available
  for the gsutil mv command (except for the -R flag, which is implied by the
  gsutil mv command). Please see the OPTIONS sections of "gsutil help cp"
  for more information.

""")


class MvCommand(Command):
  """Implementation of gsutil mv command.

     Note that there is no atomic rename operation - this command is simply
     a shorthand for 'cp' followed by 'rm'.
  """

  # Command specification. See base class for documentation.
  command_spec = Command.CreateCommandSpec(
      'mv',
      command_name_aliases=['move', 'ren', 'rename'],
      usage_synopsis=_SYNOPSIS,
      min_args=1,
      max_args=NO_MAX,
      # Flags for mv are passed through to cp.
      supported_sub_args=CP_SUB_ARGS,
      file_url_ok=True,
      provider_url_ok=False,
      urls_start_arg=0,
      gs_api_support=[ApiSelector.XML, ApiSelector.JSON],
      gs_default_api=ApiSelector.JSON,
      argparse_arguments=[
          CommandArgument.MakeZeroOrMoreCloudOrFileURLsArgument()
      ],
  )
  # Help specification. See help_provider.py for documentation.
  help_spec = Command.HelpSpec(
      help_name='mv',
      help_name_aliases=['move', 'rename'],
      help_type='command_help',
      help_one_line_summary='Move/rename objects',
      help_text=_DETAILED_HELP_TEXT,
      subcommand_help_text={},
  )

  def RunCommand(self):
    """Command entry point for the mv command."""
    # Check each source arg up, refusing to delete a bucket src URL (force users
    # to explicitly do that as a separate operation).
    for arg_to_check in self.args[0:-1]:
      url = StorageUrlFromString(arg_to_check)
      if url.IsCloudUrl() and (url.IsBucket() or url.IsProvider()):
        raise CommandException('You cannot move a source bucket using the mv '
                               'command. If you meant to move\nall objects in '
                               'the bucket, you can use a command like:\n'
                               '\tgsutil mv %s/* %s' %
                               (arg_to_check, self.args[-1]))

    # Insert command-line opts in front of args so they'll be picked up by cp
    # and rm commands (e.g., for -p option). Use undocumented (internal
    # use-only) cp -M option, which causes each original object to be deleted
    # after successfully copying to its destination, and also causes naming
    # behavior consistent with Unix mv naming behavior (see comments in
    # ConstructDstUrl).
    unparsed_args = ['-M']
    if self.recursion_requested:
      unparsed_args.append('-R')
    unparsed_args.extend(self.unparsed_args)
    self.command_runner.RunNamedCommand(
        'cp',
        args=unparsed_args,
        headers=self.headers,
        debug=self.debug,
        trace_token=self.trace_token,
        user_project=self.user_project,
        parallel_operations=self.parallel_operations)

    return 0
